"""The click-based vulnerability analysis commands.
"""

import csv
import functools
import logging
import os

import click

from clapper.click import verbosity_option
from click.types import FLOAT
from numpy import random

from bob.bio.base.score.load import split_csv_vuln
from bob.io.base import create_directories_safe
from bob.measure.script import common_options

from . import vuln_figure as figure

logger = logging.getLogger(__name__)


def vuln_plot_options(
    docstring,
    plot_output_default="vuln_plot.pdf",
    legend_loc_default="best",
    axes_lim_default=None,
    figsize_default="4,3",
    force_eval=False,
    x_label_rotation_default=0,
):
    def custom_options_command(func):
        func.__doc__ = docstring

        def eval_if_not_forced(force_eval):
            def decorator(f):
                if not force_eval:
                    return common_options.eval_option()(
                        common_options.sep_dev_eval_option()(
                            common_options.hide_dev_option()(f)
                        )
                    )
                else:
                    return f

            return decorator

        @click.command()
        @common_options.scores_argument(
            min_arg=1, force_eval=force_eval, nargs=-1
        )
        @eval_if_not_forced(force_eval)
        @common_options.legends_option()
        @common_options.no_legend_option()
        @common_options.legend_ncols_option()
        @common_options.legend_loc_option(dflt=legend_loc_default)
        @common_options.output_plot_file_option(default_out=plot_output_default)
        @common_options.lines_at_option(dflt=" ")
        @common_options.axes_val_option(dflt=axes_lim_default)
        @common_options.x_rotation_option(dflt=x_label_rotation_default)
        @common_options.x_label_option()
        @common_options.y_label_option()
        @common_options.points_curve_option()
        @common_options.const_layout_option()
        @common_options.figsize_option(dflt=figsize_default)
        @common_options.style_option()
        @common_options.linestyles_option()
        @common_options.alpha_option()
        @verbosity_option(logger=logger)
        @click.pass_context
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return custom_options_command


def real_data_option(**kwargs):
    """Option to choose if input data is real or generated"""
    return common_options.bool_option(
        name="real-data",
        short_name="R",
        desc="If False, will annotate the plots hypothetically, instead "
        "of with real data values of the calculated error rates.",
        dflt=True,
        **kwargs,
    )


def fnmr_at_option(dflt=" ", **kwargs):
    """Get option to draw const FNMR lines"""
    return common_options.list_float_option(
        name="fnmr",
        short_name="fnmr",
        desc="If given, draw horizontal lines at the given FNMR position. "
        "Your values must be separated with a comma (,) without space. "
        "This option works in ROC and DET curves.",
        nitems=None,
        dflt=dflt,
        **kwargs,
    )


def gen_score_distr(
    mean_gen,
    mean_zei,
    mean_pa,
    sigma_gen=1,
    sigma_zei=1,
    sigma_pa=1,
    num_gen=5000,
    num_zei=5000,
    num_pa=5000,
):
    # initialise the random number generator
    mt = random.RandomState(0)

    genuine_scores = mt.normal(loc=mean_gen, scale=sigma_gen, size=num_gen)
    zei_scores = mt.normal(loc=mean_zei, scale=sigma_zei, size=num_zei)
    pa_scores = mt.normal(loc=mean_pa, scale=sigma_pa, size=num_pa)

    return genuine_scores, zei_scores, pa_scores


def write_scores_to_file(neg_licit, pos_licit, spoof, filename):
    """Writes score distributions into a CSV score file. For the format of
      the score files, please refer to Bob's documentation.

    Parameters
    ----------
    neg : array_like
        Scores for negative samples.
    pos : array_like
        Scores for positive samples.
    filename : str
        The path to write the score to.
    """
    logger.info(f"Creating score file '{filename}'")
    create_directories_safe(os.path.dirname(filename))
    with open(filename, "wt") as f:
        csv_writer = csv.writer(f)
        # Write the header
        csv_writer.writerow(
            [
                "bio_ref_subject_id",
                "probe_subject_id",
                "probe_key",
                "probe_attack_type",
                "score",
            ]
        )
        for score in neg_licit:
            csv_writer.writerow(["x", "y", "0", None, score])
        for score in pos_licit:
            csv_writer.writerow(["x", "x", "0", None, score])
        for score in spoof:
            csv_writer.writerow(["x", "y", "0", "pai", score])


@click.command()
@click.argument("outdir")
@click.option("-mg", "--mean-gen", default=7, type=FLOAT, show_default=True)
@click.option("-mz", "--mean-zei", default=3, type=FLOAT, show_default=True)
@click.option("-mp", "--mean-pa", default=5, type=FLOAT, show_default=True)
@verbosity_option(logger=logger)
def gen(outdir, mean_gen, mean_zei, mean_pa, **kwargs):
    """Generate random scores.
    Generates random scores for three types of verification attempts:
    genuine users, zero-effort impostors and spoofing attacks and writes them
    into CSV score files for so called licit and spoof scenario. The
    scores are generated using Gaussian distribution whose mean is an input
    parameter. The generated scores can be used as hypothetical datasets.
    """
    # Generate the data
    genuine_dev, zei_dev, pa_dev = gen_score_distr(mean_gen, mean_zei, mean_pa)
    genuine_eval, zei_eval, pa_eval = gen_score_distr(
        mean_gen, mean_zei, mean_pa
    )

    # Write the data into files
    write_scores_to_file(
        zei_dev, genuine_dev, pa_dev, os.path.join(outdir, "scores-dev.csv")
    )
    write_scores_to_file(
        zei_eval, genuine_eval, pa_eval, os.path.join(outdir, "scores-eval.csv")
    )


@common_options.metrics_command(
    docstring="""Extracts different statistical values from scores distributions

    Prints a table that contains different metrics to assess the performance of a
    biometric system against zero-effort impostor and presentation attacks.

    The CSV score files must contain an `attack-type` column, in addition to the
    "regular" biometric scores columns (`bio_ref_subject_id`,
    `probe_subject_id`, and `score`).

    Examples:

        $ bob vuln metrics -v scores-dev.csv

        $ bob vuln metrics -v -e scores-{dev,eval}.csv
  """
)
@common_options.cost_option()
def metrics(ctx, scores, evaluation, **kwargs):
    process = figure.Metrics(ctx, scores, evaluation, split_csv_vuln)
    process.run()


@vuln_plot_options(
    docstring="""Plots the ROC for vulnerability analysis

    You need to provide 1 or 2 (with `--eval`) score
    files for each vulnerability system in this order:

    \b
    * dev scores
    * [eval scores]

    The CSV score files must contain an `attack-type` column, in addition to the
    "regular" biometric scores columns (`bio_ref_subject_id`,
    `probe_subject_id`, and `score`).

    Examples:

        $ bob vuln roc -v -o roc.pdf scores.csv

        $ bob vuln roc -v -e scores-{dev,eval}.csv
  """,
    plot_output_default="vuln_roc.pdf",
)
@common_options.title_option()
@common_options.min_far_option()
@common_options.tpr_option(dflt=True)
@common_options.semilogx_option(dflt=True)
@fnmr_at_option()
@real_data_option()
def roc(ctx, scores, evaluation, real_data, **kwargs):
    process = figure.RocVuln(
        ctx, scores, evaluation, split_csv_vuln, real_data, False
    )
    process.run()


@vuln_plot_options(
    docstring="""Plots the DET for vulnerability analysis

    You need to provide 1 or 2 (with `--eval`) score
    files for each vulnerability system in this order:

    \b
    * dev scores
    * [eval scores]

    The CSV score files must contain an `attack-type` column, in addition to the
    "regular" biometric scores columns (`bio_ref_subject_id`,
    `probe_subject_id`, and `score`).

    See :ref:`bob.bio.base.vulnerability` in the documentation for a guide on
    vulnerability analysis.

    Examples:

        $ bob vuln det -v -o det.pdf scores.csv

        $ bob vuln det -v -e scores-{dev,eval}.csv
  """,
    plot_output_default="vuln_det.pdf",
    legend_loc_default="upper-right",
    axes_lim_default="0.01,95,0.01,95",
    figsize_default="6,4",
    x_label_rotation_default=45,
)
@common_options.title_option()
@real_data_option()
@fnmr_at_option()
def det(ctx, scores, evaluation, real_data, **kwargs):
    process = figure.DetVuln(
        ctx, scores, evaluation, split_csv_vuln, real_data, False
    )
    process.run()


@vuln_plot_options(
    docstring="""Plots the EPC for vulnerability analysis

    You need to provide 2 score files for each vulnerability system in this order:

    \b
    * dev scores
    * eval scores

    The CSV score files must contain an `attack-type` column, in addition to the
    "regular" biometric scores columns (`bio_ref_subject_id`,
    `probe_subject_id`, and `score`).

    See :ref:`bob.bio.base.vulnerability` in the documentation for a guide on
    vulnerability analysis.

    Examples:

        $ bob vuln epc -v scores-dev.csv scores-eval.csv

        $ bob vuln epc -v -o epc.pdf scores-{dev,eval}.csv
  """,
    plot_output_default="vuln_epc.pdf",
    force_eval=True,
    legend_loc_default="upper-center",
)
@common_options.title_option()
@common_options.bool_option(
    "iapmr", "I", "Whether to plot the IAPMR related lines or not.", True
)
def epc(ctx, scores, **kwargs):
    process = figure.Epc(ctx, scores, True, split_csv_vuln)
    process.run()


@vuln_plot_options(
    docstring="""Plots the EPSC for vulnerability analysis

    Plots the Expected Performance Spoofing Curve.

    Note that when using 3D plots with option ``--three-d``, you cannot plot
    both WER and IAPMR on the same figure (which is possible in 2D).

    You need to provide 2 score files for each vulnerability system in this order:

    \b
    * dev scores
    * eval scores

    The CSV score files must contain an `attack-type` column, in addition to the
    "regular" biometric scores columns (`bio_ref_subject_id`,
    `probe_subject_id`, and `score`).

    See :ref:`bob.bio.base.vulnerability` in the documentation for a guide on
    vulnerability analysis.

    Examples:

        $ bob vuln epsc -v scores-dev.csv scores-eval.csv

        $ bob vuln epsc -v -o epsc.pdf scores-{dev,eval}.csv

        $ bob vuln epsc -v -D -o epsc_3D.pdf scores-{dev,eval}.csv
  """,
    plot_output_default="vuln_epc.pdf",
    force_eval=True,
    figsize_default="5,3",
)
@common_options.titles_option()
@common_options.bool_option(
    "wer", "w", "Whether to plot the WER related lines or not.", True
)
@common_options.bool_option(
    "iapmr", "I", "Whether to plot the IAPMR related lines or not.", True
)
@common_options.bool_option(
    "three-d",
    "D",
    "If true, generate 3D plots. You need to turn off "
    "wer or iapmr when using this option.",
    False,
)
@click.option(
    "-c",
    "--criteria",
    default="eer",
    show_default=True,
    help="Criteria for threshold selection",
    type=click.Choice(("eer", "min-hter")),
)
@click.option(
    "-vp",
    "--var-param",
    default="omega",
    show_default=True,
    help="Name of the varying parameter",
    type=click.Choice(("omega", "beta")),
)
@common_options.list_float_option(
    name="fixed-params",
    short_name="fp",
    dflt="0.5",
    desc="Values of the fixed parameter, separated by commas",
)
@click.option(
    "-s",
    "--sampling",
    default=5,
    show_default=True,
    help="Sampling of the EPSC 3D surface",
    type=click.INT,
)
def epsc(ctx, scores, criteria, var_param, three_d, sampling, **kwargs):
    if three_d:
        if ctx.meta["wer"] and ctx.meta["iapmr"]:
            logger.info(
                "Cannot plot both WER and IAPMR in 3D. Will turn IAPMR off."
            )
            ctx.meta["iapmr"] = False
        ctx.meta["sampling"] = sampling
        process = figure.Epsc3D(
            ctx, scores, split_csv_vuln, criteria, var_param, **kwargs
        )
    else:
        process = figure.Epsc(
            ctx, scores, split_csv_vuln, criteria, var_param, **kwargs
        )
    process.run()


@vuln_plot_options(
    docstring="""Vulnerability analysis score distribution histograms.

    Plots the histogram of score distributions. You need to provide 1 or 2 score
    files for each biometric system in this order:

    \b
    * development scores
    * [evaluation scores]

    When evaluation scores are provided, you must use the ``--eval`` option.

    The CSV score files must contain an `attack-type` column, in addition to the
    "regular" biometric scores columns (`bio_ref_subject_id`,
    `probe_subject_id`, and `score`).

    See :ref:`bob.bio.base.vulnerability` in the documentation for a guide on
    vulnerability analysis.

    When eval-scores are given, eval-scores histograms are displayed with the
    threshold line computed from dev-scores.

    Examples:

        $ bob vuln hist -v -o hist.pdf results/scores-dev.csv

        $ bob vuln hist -e -v results/scores-dev.csv results/scores-eval.csv

        $ bob vuln hist -e -v results/scores-{dev,eval}.csv
  """,
    plot_output_default="vuln_hist.pdf",
)
@common_options.titles_option()
@common_options.n_bins_option()
@common_options.thresholds_option()
@common_options.print_filenames_option(dflt=False)
@common_options.bool_option(
    "iapmr-line", "I", "Whether to plot the IAPMR related lines or not.", True
)
@real_data_option()
@common_options.subplot_option()
@common_options.criterion_option()
def hist(ctx, scores, evaluation, **kwargs):
    process = figure.HistVuln(ctx, scores, evaluation, split_csv_vuln)
    process.run()


@vuln_plot_options(
    docstring="""Plots the FMR vs IAPMR for vulnerability analysis

    You need to provide 1 or 2 (with `--eval`) score
    files for each vulnerability system in this order:

    \b
    * dev scores
    * [eval scores]

    The CSV score files must contain an `attack-type` column, in addition to the
    "regular" biometric scores columns (`bio_ref_subject_id`,
    `probe_subject_id`, and `score`).

    Examples:

        $ bob vuln fmr_iapmr -v -o fmr_iapmr.pdf scores.csv

        $ bob vuln fmr_iapmr -v -e scores-{dev,eval}.csv
  """,
    plot_output_default="vuln_roc.pdf",
    force_eval=True,
)
@common_options.title_option()
@common_options.semilogx_option()
def fmr_iapmr(ctx, scores, **kwargs):
    process = figure.FmrIapmr(ctx, scores, True, split_csv_vuln)
    process.run()


@common_options.evaluate_command(
    common_options.EVALUATE_HELP.format(
        score_format=(
            "Files must be in CSV format, with the `bio_ref_subject_id`, "
            "`probe_references_id`, `score`, and `attack_type` columns."
        ),
        command="bob vuln evaluate",
    ),
    criteria=("eer", "min-hter", "far"),
)
def evaluate(ctx, scores, evaluation, **kwargs):
    # open_mode is always 'write' in this command.
    ctx.meta["open_mode"] = "w"
    criterion = ctx.meta.get("criterion")
    if criterion is not None:
        click.echo(f"Computing metrics with {criterion}...")
        ctx.invoke(metrics, scores=scores, evaluation=evaluation)
        if ctx.meta.get("log") is not None:
            click.echo(f"[metrics] => {ctx.meta['log']}")

    ctx.meta["lines_at"] = None

    # Avoid closing pdf file before all figures are plotted
    ctx.meta["closef"] = False
    if evaluation:
        click.echo("Starting evaluate with dev and eval scores...")
    else:
        click.echo("Starting evaluate with dev scores only...")
    click.echo("Plotting FMR vs IAPMR for bob vuln evaluate...")
    ctx.forward(fmr_iapmr)  # uses class defaults plot settings
    click.echo("Plotting ROC for bob vuln evaluate...")
    ctx.forward(roc)  # uses class defaults plot settings
    click.echo("Plotting DET for bob vuln evaluate...")
    ctx.forward(det)  # uses class defaults plot settings
    if evaluation:
        click.echo("Plotting EPSC for bob vuln evaluate...")
        ctx.forward(epsc)  # uses class defaults plot settings
    # Mark the last plot to close the output file
    ctx.meta["closef"] = True
    click.echo("Plotting score histograms for bob vuln evaluate...")
    ctx.forward(hist)
    click.echo("Evaluate successfully completed!")
    click.echo(f"[plots] => {ctx.meta['output']}")



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Andre Anjos <andre.anjos@idiap.ch>
# Wed 11 Dec 15:14:08 2013 CET
#
# Copyright (C) 2011-2013 Idiap Research Institute, Martigny, Switzerland

"""Tests the IO functionality of bob.bio.base.score."""

import tempfile

import numpy
import pkg_resources

from bob.bio.base import score


def test_load_scores():
    # This function tests the IO functionality of loading score files in
    # different ways

    load_functions = {"4col": score.four_column, "5col": score.five_column}
    cols = {"4col": 4, "5col": 5}

    for variant in cols:
        normal_score_file = pkg_resources.resource_filename(
            __name__, "data/dev-%s.txt" % variant
        )
        normal_scores = list(load_functions[variant](normal_score_file))

        assert len(normal_scores) == 910
        assert all(len(s) == cols[variant] for s in normal_scores)

        # read the compressed score file
        compressed_score_file = pkg_resources.resource_filename(
            __name__, "data/dev-%s.tar.gz" % variant
        )
        compressed_scores = list(load_functions[variant](compressed_score_file))

        assert len(compressed_scores) == len(normal_scores)
        assert all(len(c) == cols[variant] for c in compressed_scores)
        assert all(
            c[i] == s[i]
            for c, s in zip(compressed_scores, normal_scores)
            for i in range(cols[variant])
        )

        # Use auto-estimated score file contents
        # read score file in normal way
        normal_scores = list(score.scores(normal_score_file))

        assert len(normal_scores) == 910
        assert all(len(s) == cols[variant] for s in normal_scores)

        # read the compressed score file
        compressed_scores = list(score.scores(compressed_score_file))

        assert len(compressed_scores) == len(normal_scores)
        assert all(len(c) == cols[variant] for c in compressed_scores)
        assert all(
            c[i] == s[i]
            for c, s in zip(compressed_scores, normal_scores)
            for i in range(cols[variant])
        )


def test_split_vuln_scores():
    """Tests that vulnerability score files are loaded correctly"""
    score_file = pkg_resources.resource_filename(
        __name__, "data/vuln/scores-dev.csv"
    )
    split_scores = score.split_csv_vuln(score_file)
    assert all(
        key in split_scores for key in ("licit_neg", "licit_pos", "spoof")
    )
    assert all(
        isinstance(scores, numpy.ndarray) for scores in split_scores.values()
    )
    assert all(len(scores) == 5000 for scores in split_scores.values())


def test_split_scores():
    # This function tests the IO functionality of loading score files in
    # different ways

    split_functions = {
        "4col": score.split_four_column,
        "5col": score.split_five_column,
    }
    cols = {"4col": 4, "5col": 5}

    for variant in cols:
        # read score file in normal way
        normal_score_file = pkg_resources.resource_filename(
            __name__, "data/dev-%s.txt" % variant
        )
        negatives, positives = split_functions[variant](normal_score_file)

        assert len(negatives) == 520, len(negatives)
        assert len(positives) == 390, len(positives)

        # read the compressed score file
        compressed_score_file = pkg_resources.resource_filename(
            __name__, "data/dev-%s.tar.gz" % variant
        )
        negatives, positives = split_functions[variant](compressed_score_file)

        assert len(negatives) == 520, len(negatives)
        assert len(positives) == 390, len(positives)

        # Use auto-estimated score file contents
        # read score file in normal way
        negatives, positives = score.split(normal_score_file)

        assert len(negatives) == 520, len(negatives)
        assert len(positives) == 390, len(positives)

        # read the compressed score file
        negatives, positives = score.split(compressed_score_file)

        assert len(negatives) == 520, len(negatives)
        assert len(positives) == 390, len(positives)


def test_load_score():
    # This function tests the IO functionality of loading score files in
    # different ways

    cols = {"4col": 4, "5col": 5}

    for variant in cols:
        compressed_score_file = pkg_resources.resource_filename(
            __name__, "data/dev-%s.tar.gz" % variant
        )
        normal_score_file = pkg_resources.resource_filename(
            __name__, "data/dev-%s.txt" % variant
        )
        normal_scores = score.load_score(normal_score_file, cols[variant])

        assert len(normal_scores) == 910
        assert len(normal_scores.dtype) == cols[variant]

        # read the compressed score file
        compressed_score_file = pkg_resources.resource_filename(
            __name__, "data/dev-%s.tar.gz" % variant
        )
        compressed_scores = score.load_score(
            compressed_score_file, cols[variant]
        )

        assert len(compressed_scores) == len(normal_scores)
        assert len(compressed_scores.dtype) == cols[variant]
        for name in normal_scores.dtype.names:
            assert all(normal_scores[name] == compressed_scores[name])

        # test minimal loading
        minimal_scores = score.load_score(normal_score_file, minimal=True)
        assert len(minimal_scores) == 910
        assert len(minimal_scores.dtype) == 3
        assert minimal_scores.dtype.names == ("claimed_id", "real_id", "score")


def test_dump_score():
    # This function tests the IO functionality of dumping score files

    cols = {"4col": 4, "5col": 5}

    for variant in cols:
        # read score file
        normal_score_file = pkg_resources.resource_filename(
            __name__, "data/dev-%s.txt" % variant
        )
        normal_scores = score.load_score(normal_score_file, cols[variant])

        with tempfile.TemporaryFile() as f:
            score.dump_score(f, normal_scores)
            f.seek(0)
            loaded_scores = score.load_score(f, cols[variant])

        for name in normal_scores.dtype.names:
            assert all(normal_scores[name] == loaded_scores[name])


def _check_binary_identical(name1, name2):
    # see: http://www.peterbe.com/plog/using-md5-to-check-equality-between-files
    from hashlib import md5

    # tests if two files are binary identical
    with open(name1, "rb") as f1:
        with open(name2, "rb") as f2:
            assert md5(f1.read()).digest() == md5(f2.read()).digest()



#!/usr/bin/env python
"""
ATNT database implementation
"""

from pathlib import Path

from sklearn.pipeline import make_pipeline

import bob.io.base

from bob.bio.base.database.utils import download_file, md5_hash

from . import CSVDatabase, FileSampleLoader


class AtntBioDatabase(CSVDatabase):
    """
    The AT&T (aka ORL) database of faces
    (http://www.cl.cam.ac.uk/research/dtg/attarchive/facedatabase.html). This
    class defines a simple protocol for training, enrollment and probe by
    splitting the few images of the database in a reasonable manner. Due to the
    small size of the database, there is only a 'dev' group, and I did not
    define an 'eval' group.
    """

    category = "base"
    dataset_protocols_urls = [
        "https://www.idiap.ch/software/bob/databases/latest/base/atnt-f529acef.tar.gz",
        "http://www.idiap.ch/software/bob/databases/latest/base/atnt-f529acef.tar.gz",
    ]
    dataset_protocols_checksum = "f529acef"

    def __init__(
        self,
        protocol="idiap_protocol",
        dataset_original_directory=None,
        **kwargs,
    ):
        """Custom init to download the raw data files."""

        # Download the raw data (or use a cached local version)
        if dataset_original_directory is None:
            path = download_file(
                urls=[
                    "http://www.idiap.ch/software/bob/data/bob/att_faces.zip"
                ],
                destination_sub_directory="datasets/atnt",
                destination_filename="atnt_faces.zip",
                checksum="6efb25cb0d40755e9492b9c012e3348d",
                checksum_fct=md5_hash,
                extract=True,
            )
            dataset_original_directory = Path(path).as_posix()

        super().__init__(
            name="atnt",
            protocol=protocol,
            transformer=make_pipeline(
                FileSampleLoader(
                    data_loader=bob.io.base.load,
                    dataset_original_directory=dataset_original_directory,
                    extension=".pgm",
                ),
            ),
            **kwargs,
        )
        # just expose original_directory for backward compatibility of tests
        self.original_directory = dataset_original_directory
        self.original_extension = ".pgm"

    # define an objects method for compatibility with the old tests
    def objects(
        self, model_ids=None, groups=None, purposes=None, protocol=None
    ):
        samples = []

        if groups is None:
            groups = self.groups()

        if "train" in groups or "world" in groups:
            samples += self.background_model_samples()

        if purposes is None:
            purposes = ("enroll", "probe")

        if "enroll" in purposes:
            samples += self.references()

        if "probe" in purposes:
            samples += self.probes()

        if model_ids:
            samples = [s for s in samples if s.template_id in model_ids]

        # create the old attributes
        for s in samples:
            s.client_id = s.template_id
            s.path = s.id = s.key

        return samples



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# @author: Manuel Guenther <Manuel.Guenther@idiap.ch>
# @date: Thu Oct 25 10:05:55 CEST 2012

from __future__ import print_function

import logging
import os

import clapper.config
import pkg_resources

logger = logging.getLogger("bob.bio.base")


valid_keywords = (
    "database",
    "preprocessor",
    "extractor",
    "algorithm",
    "grid",
    "client",
    "config",
    "annotator",
    "pipeline",
)
"""Keywords for which resources are defined."""


def _collect_config(paths):
    """Collect all python file resources into a module

    This function recursively loads python modules (in a Python 3-compatible way)
    so the last loaded module corresponds to the final state of the loading. In
    this way, we load the first file, resolve its symbols, overwrite with the
    second file and so on. We return a temporarily created module containing all
    resolved variables, respecting the input order.


    Parameters:

      paths : [str]
        A list of resources, modules or files (in order) to collect resources from


    Returns: module

      A valid Python module you can use to configure your tool

    """

    return clapper.config.load(paths, entry_point_group="bob.bio.config")


def read_config_file(filenames, keyword=None):
    """Use this function to read the given configuration file.
    If a keyword is specified, only the configuration according to this keyword is returned.
    Otherwise a dictionary of the configurations read from the configuration file is returned.

    **Parameters:**

    filenames : [str]
      A list (pontentially empty) of configuration files or resources to read
      running options from

    keyword : str or ``None``
      If specified, only the contents of the variable with the given name is returned.
      If ``None``, the whole configuration is returned (a local namespace)

    **Returns:**

    config : object or namespace
      If ``keyword`` is specified, the object inside the configuration with the given name is returned.
      Otherwise, the whole configuration is returned (as a local namespace).
    """

    if not filenames:
        raise RuntimeError(
            "At least one configuration file, resource or "
            "module name must be passed"
        )

    config = _collect_config(filenames)

    if not keyword:
        return config

    if not hasattr(config, keyword):
        raise ImportError(
            "The desired keyword '%s' does not exist in any of "
            "your configuration files: %s" % (keyword, ", ".join(filenames))
        )

    return getattr(config, keyword)


def _get_entry_points(keyword, strip=[], package_prefix="bob.bio."):
    """Returns the list of entry points for registered resources with the given keyword."""
    return [
        entry_point
        for entry_point in pkg_resources.iter_entry_points(
            package_prefix + keyword
        )
        if not entry_point.name.startswith(tuple(strip))
    ]


def load_resource(
    resource,
    keyword,
    imports=["bob.bio.base"],
    package_prefix="bob.bio.",
    preferred_package=None,
):
    """Loads the given resource that is registered with the given keyword.
    The resource can be:

    1. a resource as defined in the setup.py
    2. a configuration file
    3. a string defining the construction of an object. If imports are required for the construction of this object, they can be given as list of strings.

    **Parameters:**

    resource : str
      Any string interpretable as a resource (see above).

    keyword : str
      A valid resource keyword, can be one of :any:`bob.bio.base.utils.resources.valid_keywords`.

    imports : [str]
      A list of strings defining which modules to import, when constructing new objects (option 3).

    package_prefix : str
      Package namespace, in which we search for entry points, e.g., ``bob.bio``.

    preferred_package : str or ``None``
      When several resources with the same name are found in different packages (e.g., in different ``bob.bio`` or other packages), this specifies the preferred package to load the resource from.
      If not specified, the extension that is **not** from ``bob.bio`` is selected.

    **Returns:**

    resource : object
      The resulting resource object is returned, either read from file or resource, or created newly.
    """

    # first, look if the resource is a file name
    if os.path.isfile(resource):
        return read_config_file([resource], keyword)

    if keyword not in valid_keywords:
        logger.warning(
            "The given keyword '%s' is not valid. Please use one of %s!",
            keyword,
            valid_keywords,
        )

    # now, we check if the resource is registered as an entry point in the resource files
    entry_points = [
        entry_point
        for entry_point in _get_entry_points(
            keyword, package_prefix=package_prefix
        )
        if entry_point.name == resource
    ]

    if len(entry_points):
        if len(entry_points) == 1:
            return entry_points[0].load()
        else:
            # TODO: extract current package name and use this one, if possible

            # Now: check if there are only two entry points, and one is from the bob.bio.base, then use the other one
            index = -1
            if preferred_package is not None:
                for i, p in enumerate(entry_points):
                    if p.dist.project_name == preferred_package:
                        index = i
                        break

            if index == -1:
                # by default, use the first one that is not from bob.bio
                for i, p in enumerate(entry_points):
                    if not p.dist.project_name.startswith(package_prefix):
                        index = i
                        break

            if index != -1:
                logger.debug(
                    "RESOURCES: Using the resource '%s' from '%s', and ignoring the one from '%s'",
                    resource,
                    entry_points[index].module_name,
                    entry_points[1 - index].module_name,
                )
                return entry_points[index].load()
            else:
                logger.warn(
                    "Under the desired name '%s', there are multiple entry points defined, we return the first one: %s",
                    resource,
                    [entry_point.module_name for entry_point in entry_points],
                )
                return entry_points[0].load()

    # if the resource is neither a config file nor an entry point,
    # just execute it as a command
    try:
        # first, execute all import commands that are required
        for i in imports:
            exec("import %s" % i)
        # now, evaluate the resource (re-evaluate if the resource is still a string)
        while isinstance(resource, str):
            resource = eval(resource)
        return resource

    except Exception as e:
        raise ImportError(
            "The given command line option '%s' is neither a resource for a '%s', nor an existing configuration file, nor could be interpreted as a command"
            % (resource, keyword)
        ) from e


def extensions(keywords=valid_keywords, package_prefix="bob.bio."):
    """extensions(keywords=valid_keywords, package_prefix='bob.bio.') -> extensions

    Returns a list of packages that define extensions using the given keywords.

    **Parameters:**

    keywords : [str]
      A list of keywords to load entry points for.
      Defaults to all :any:`bob.bio.base.utils.resources.valid_keywords`.

    package_prefix : str
      Package namespace, in which we search for entry points, e.g., ``bob.bio``.
    """
    entry_points = [
        entry_point
        for keyword in keywords
        for entry_point in _get_entry_points(
            keyword, package_prefix=package_prefix
        )
    ]
    return sorted(
        list(set(entry_point.dist.project_name for entry_point in entry_points))
    )


def resource_keys(
    keyword, exclude_packages=[], package_prefix="bob.bio.", strip=["dummy"]
):
    """Reads and returns all resources that are registered with the given keyword.
    Entry points from the given ``exclude_packages`` are ignored."""
    ret_list = [
        entry_point.name
        for entry_point in _get_entry_points(
            keyword, strip=strip, package_prefix=package_prefix
        )
        if entry_point.dist.project_name not in exclude_packages
    ]
    return sorted(ret_list)


def list_resources(
    keyword,
    strip=["dummy"],
    package_prefix="bob.bio.",
    verbose=False,
    packages=None,
):
    """Returns a string containing a detailed list of resources that are registered with the given keyword."""
    if keyword not in valid_keywords:
        raise ValueError(
            "The given keyword '%s' is not valid. Please use one of %s!"
            % (str(keyword), str(valid_keywords))
        )

    entry_points = _get_entry_points(
        keyword, strip, package_prefix=package_prefix
    )
    last_dist = None
    retval = ""
    length = (
        max(len(entry_point.name) for entry_point in entry_points)
        if entry_points
        else 1
    )

    if packages is not None:
        entry_points = [
            entry_point
            for entry_point in entry_points
            if entry_point.dist.project_name in packages
        ]

    for entry_point in sorted(
        entry_points, key=lambda p: (p.dist.project_name, p.name)
    ):
        if last_dist != str(entry_point.dist):
            retval += "\n- %s @ %s: \n" % (
                str(entry_point.dist),
                str(entry_point.dist.location),
            )
            last_dist = str(entry_point.dist)

        if len(entry_point.attrs):
            retval += "  + %s --> %s: %s\n" % (
                entry_point.name + " " * (length - len(entry_point.name)),
                entry_point.module_name,
                entry_point.attrs[0],
            )
        else:
            retval += "  + %s --> %s\n" % (
                entry_point.name + " " * (length - len(entry_point.name)),
                entry_point.module_name,
            )
        if verbose:
            retval += "    ==> " + str(entry_point.load()) + "\n\n"

    return retval


def database_directories(
    strip=["dummy"], replacements=None, package_prefix="bob.bio."
):
    """Returns a dictionary of original directories for all registered databases."""
    entry_points = _get_entry_points(
        "database", strip, package_prefix=package_prefix
    )

    dirs = {}
    for entry_point in sorted(
        entry_points, key=lambda entry_point: entry_point.name
    ):
        try:
            db = load_resource(entry_point.name, "database")
            db.replace_directories(replacements)
            dirs[entry_point.name] = [db.original_directory]
            if db.annotation_directory is not None:
                dirs[entry_point.name].append(db.annotation_directory)
        except (AttributeError, ValueError, ImportError):
            pass

    return dirs


def get_resource_filename(resource_name, group):
    """
    Get the file name of a resource.


    Parameters
    ----------
        resource_name: str
            Name of the resource to be searched

        group: str
            Entry point group

    Return
    ------
        filename: str
            The entrypoint file name

    """

    # Check if it's already a path
    if os.path.exists(resource_name):
        return resource_name

    # If it's a resource get the path of this resource
    resources = [r for r in pkg_resources.iter_entry_points(group)]

    # if resource_name not in [r.name for r in resources]:
    #    raise ValueError(f"Resource not found: `{resource_name}`")

    for r in resources:
        if r.name == resource_name:
            resource = r
            break
    else:
        raise ValueError(f"Resource not found: `{resource_name}`")

    # TODO: This get the root path only
    #        I don't know how to get the filename
    return (
        pkg_resources.resource_filename(
            resource.module_name, resource.module_name.split(".")[-1]
        )
        + ".py"
    )



"""Interface between the lower level GMM classes and the Algorithm Transformer.

Implements the enroll and score methods using the low level GMM implementation.

This adds the notions of models, probes, enrollment, and scores to GMM.
"""


import copy
import logging

from typing import Callable, Union

import dask.array as da
import numpy as np

from h5py import File as HDF5File

from bob.bio.base.pipelines import BioAlgorithm
from bob.learn.em import GMMMachine, GMMStats, linear_scoring

logger = logging.getLogger(__name__)


def check_data_dim(data, expected_ndim):
    """Stacks the features into a matrix of shape (n_samples, n_features) or
    labels into shape of (n_samples,) if the input data is not like that already

    Parameters
    ----------
    data : array-like
        features or labels
    expected_ndim : int
        expected number of dimensions of the data

    Returns
    -------
    stacked_data : array-like
        stacked features or labels if needed
    """
    if expected_ndim not in (1, 2):
        raise ValueError(
            f"expected_ndim must be 1 or 2 but got {expected_ndim}"
        )

    if expected_ndim == 1:
        stack_function = np.concatenate
    else:
        stack_function = np.vstack

    if data[0].ndim == expected_ndim:
        return stack_function(data)

    return data


class GMM(GMMMachine, BioAlgorithm):
    """Algorithm for computing UBM and Gaussian Mixture Models of the features.

    Features must be normalized to zero mean and unit standard deviation.

    Models are MAP GMM machines trained from a UBM on the enrollment feature set.

    The UBM is a ML GMM machine trained on the training feature set.

    Probes are GMM statistics of features projected on the UBM.
    """

    def __init__(
        self,
        # parameters for the GMM
        n_gaussians: int,
        # parameters of UBM training
        k_means_trainer=None,
        max_fitting_steps: int = 25,  # Maximum number of iterations for GMM Training
        convergence_threshold: float = 5e-4,  # Threshold to end the ML training
        mean_var_update_threshold: float = 5e-4,  # Minimum value that a variance can reach
        update_means: bool = True,
        update_variances: bool = True,
        update_weights: bool = True,
        # parameters of the GMM enrollment (MAP)
        enroll_iterations: int = 1,
        enroll_update_means: bool = True,
        enroll_update_variances: bool = False,
        enroll_update_weights: bool = False,
        enroll_relevance_factor: Union[float, None] = 4,
        enroll_alpha: float = 0.5,
        # scoring
        scoring_function: Callable = linear_scoring,
        # RNG
        random_state: int = 5489,
        # other
        return_stats_in_transform: bool = False,
        **kwargs,
    ):
        """Initializes the local UBM-GMM tool chain.

        Parameters
        ----------
        n_gaussians
            The number of Gaussians used in the UBM and the models.
        kmeans_trainer
            The kmeans machine used to train and initialize the UBM.
        kmeans_init_iterations
            Number of iterations used for setting the k-means initial centroids.
            if None, will use the same as kmeans_training_iterations.
        kmeans_oversampling_factor
            Oversampling factor used by k-means initializer.
        max_fitting_steps
            Number of e-m iterations for training the UBM.
        convergence_threshold
            Convergence threshold to halt the GMM training early.
        mean_var_update_threshold
            Minimum value a variance of the Gaussians can reach.
        update_weights
            Decides wether the weights of the Gaussians are updated while training.
        update_means
            Decides wether the means of the Gaussians are updated while training.
        update_variances
            Decides wether the variances of the Gaussians are updated while training.
        enroll_iterations
            Number of iterations for the MAP GMM used for enrollment.
        enroll_update_weights
            Decides wether the weights of the Gaussians are updated while enrolling.
        enroll_update_means
            Decides wether the means of the Gaussians are updated while enrolling.
        enroll_update_variances
            Decides wether the variances of the Gaussians are updated while enrolling.
        enroll_relevance_factor
            For enrollment: MAP relevance factor as described in Reynolds paper.
            If None, will not apply Reynolds adaptation.
        enroll_alpha
            For enrollment: MAP adaptation coefficient.
        random_state
            Seed for the random number generation.
        scoring_function
            Function returning a score from a model, a UBM, and a probe.
        """
        super().__init__(
            n_gaussians=n_gaussians,
            trainer="ml",
            max_fitting_steps=max_fitting_steps,
            convergence_threshold=convergence_threshold,
            update_means=update_means,
            update_variances=update_variances,
            update_weights=update_weights,
            mean_var_update_threshold=mean_var_update_threshold,
            k_means_trainer=k_means_trainer,
            random_state=random_state,
            **kwargs,
        )

        self.enroll_relevance_factor = enroll_relevance_factor
        self.enroll_alpha = enroll_alpha
        self.enroll_iterations = enroll_iterations
        self.enroll_update_means = enroll_update_means
        self.enroll_update_weights = enroll_update_weights
        self.enroll_update_variances = enroll_update_variances
        self.scoring_function = scoring_function
        self.return_stats_in_transform = return_stats_in_transform

    def save_model(self, ubm_file):
        """Saves the projector (UBM) to file."""
        super().save(ubm_file)

    def load_model(self, ubm_file):
        """Loads the projector (UBM) from a file."""
        super().load(ubm_file)

    def project(self, array):
        """Computes GMM statistics against a UBM, given a 2D array of feature vectors

        This is applied to the probes before scoring.
        """
        array = check_data_dim(array, expected_ndim=2)
        logger.debug("Projecting %d feature vectors", array.shape[0])
        # Accumulates statistics
        gmm_stats = self.acc_stats(array)

        # Return the resulting statistics
        return gmm_stats

    def enroll(self, data):
        """Enrolls a GMM using MAP adaptation given a reference's feature vectors

        Returns a GMMMachine tuned from the UBM with MAP on a biometric reference data.
        """

        # if input is a list (or SampleBatch) of 2 dimensional arrays, stack them
        data = check_data_dim(data, expected_ndim=2)

        # Use the array to train a GMM and return it
        logger.info("Enrolling with %d feature vectors", data.shape[0])

        gmm = GMMMachine(
            n_gaussians=self.n_gaussians,
            trainer="map",
            ubm=copy.deepcopy(self),
            convergence_threshold=self.convergence_threshold,
            max_fitting_steps=self.enroll_iterations,
            random_state=self.random_state,
            update_means=self.enroll_update_means,
            update_variances=self.enroll_update_variances,
            update_weights=self.enroll_update_weights,
            mean_var_update_threshold=self.mean_var_update_threshold,
            map_relevance_factor=self.enroll_relevance_factor,
            map_alpha=self.enroll_alpha,
        )
        gmm.fit(data)
        return gmm

    def create_templates(self, list_of_feature_sets, enroll):
        if enroll:
            return [
                self.enroll(feature_set) for feature_set in list_of_feature_sets
            ]
        else:
            return [
                self.project(feature_set)
                for feature_set in list_of_feature_sets
            ]

    def compare(self, enroll_templates, probe_templates):
        return self.scoring_function(
            models_means=enroll_templates,
            ubm=self,
            test_stats=probe_templates,
            frame_length_normalization=True,
        )

    def read_biometric_reference(self, model_file):
        """Reads an enrolled reference model, which is a MAP GMMMachine."""
        return GMMMachine.from_hdf5(HDF5File(model_file, "r"), ubm=self)

    def write_biometric_reference(self, model: GMMMachine, model_file):
        """Write the enrolled reference (MAP GMMMachine) into a file."""
        return model.save(model_file)

    def fit(self, X, y=None, **kwargs):
        """Trains the UBM."""
        # Stack all the samples in a 2D array of features
        if isinstance(X, da.Array):
            X = X.persist()

        # if input is a list (or SampleBatch) of 2 dimensional arrays, stack them
        X = check_data_dim(X, expected_ndim=2)

        logger.debug(
            f"Training UBM machine with {self.n_gaussians} gaussians and {len(X)} samples"
        )

        super().fit(X)

        return self

    def transform(self, X, **kwargs):
        """Passthrough. Enroll applies a different transform as score."""
        # The idea would be to apply the projection in Transform (going from extracted
        # to GMMStats), but we must not apply this during the training or enrollment
        # (those require extracted data directly, not projected).
        # `project` is applied in the score function directly.
        if not self.return_stats_in_transform:
            return X
        return super().transform(X)

    @classmethod
    def custom_enrolled_save_fn(cls, data, path):
        data.save(path)

    def custom_enrolled_load_fn(self, path):
        return GMMMachine.from_hdf5(path, ubm=self)

    def _more_tags(self):
        return {
            "bob_fit_supports_dask_array": True,
            "bob_features_save_fn": GMMStats.save,
            "bob_features_load_fn": GMMStats.from_hdf5,
            "bob_enrolled_save_fn": self.custom_enrolled_save_fn,
            "bob_enrolled_load_fn": self.custom_enrolled_load_fn,
            "bob_checkpoint_features": self.return_stats_in_transform,
        }



"""Generate random scores.
"""
import csv
import logging
import os

import click
import numpy

from clapper.click import verbosity_option

logger = logging.getLogger(__name__)


def gen_score_distr(
    mean_neg,
    mean_pos,
    sigma_neg=10,
    sigma_pos=10,
    n_neg=5000,
    n_pos=5000,
    seed=0,
):
    """Generate scores from normal distributions

    Parameters
    ----------
    mean_neg : float
        Mean for negative scores
    mean_pos : float
        Mean for positive scores
    sigma_neg : float
        STDev for negative scores
    sigma_pos : float
        STDev for positive scores
    n_pos: int
        The number of positive scores generated
    n_neg: int
        The number of negative scores generated
    seed: int
        A value to initialize the Random Number generator. Giving the same
        value (or not specifying 'seed') on two different calls will generate
        the same lists of scores.

    Returns
    -------
    neg_scores : :any:`list`
        Negatives scores
    pos_scores : :any:`list`
        Positive scores
    """

    logger.debug("Initializing RNG.")
    numpy.random.seed(seed)

    logger.info(f"Generating {n_neg} negative and {n_pos} positive scores.")

    neg_scores = numpy.random.normal(loc=mean_neg, scale=sigma_neg, size=n_neg)
    pos_scores = numpy.random.normal(loc=mean_pos, scale=sigma_pos, size=n_pos)

    return neg_scores, pos_scores


def write_scores_to_file(
    neg,
    pos,
    filename,
    n_subjects=5,
    n_probes_per_subject=5,
    n_unknown_subjects=0,
    neg_unknown=None,
    to_csv=True,
    five_col=False,
    metadata={"meta0": "data0", "meta1": "data1"},
):
    """Writes score distributions

    Parameters
    ----------
    neg : :py:class:`numpy.ndarray`
        Scores for negative samples.
    pos : :py:class:`numpy.ndarray`
        Scores for positive samples.
    filename : str
        The path to write the score to.
    n_subjects: int
        Number of different subjects
    n_probes_per_subject: int
        Number of different samples used as probe for each subject
    n_unknown_subjects: int
        The number of unknown (no registered model) subjects
    neg_unknown: None or list
        The of unknown subjects scores
    to_csv: bool
        Use the CSV format, else the legacy 4 or 5 columns format.
    five_col : bool
        If 5-colum format, else 4-column
    """
    logger.debug(f"Creating result directories ('{filename}').")
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    s_subjects = ["x%d" % i for i in range(n_subjects)]

    logger.debug("Writing scores to files.")

    with open(filename, "wt") as f:
        if to_csv:
            csv_writer = csv.writer(f)
            csv_writer.writerow(
                ["bio_ref_subject_id", "probe_subject_id", "key", "score"]
                + list(metadata.keys())
            )
        # Generate one line per probe (unless "--force-count" specified)
        logger.debug("Writing positive scores.")
        for i, score in enumerate(pos):
            s_name = s_subjects[int(i / n_probes_per_subject) % n_subjects]
            s_five = " " if not five_col else " d" + s_name + " "
            probe_id = "%s_%d" % (s_name, i % n_probes_per_subject)
            if to_csv:
                csv_writer.writerow(
                    [s_name, s_name, probe_id, score] + list(metadata.values())
                )
            else:
                f.write(
                    "%s%s%s %s %f\n" % (s_name, s_five, s_name, probe_id, score)
                )

        # Generate one line per probe against each ref (unless "--force-count" specified)
        logger.debug("Writing negative scores.")
        for i, score in enumerate(neg):
            n_impostors = n_subjects - 1
            ref = s_subjects[
                int(i / n_probes_per_subject / n_impostors) % n_subjects
            ]
            impostors = [s for s in s_subjects if s != ref]  # ignore pos
            probe = impostors[int(i / n_probes_per_subject) % n_impostors]
            s_five = " " if not five_col else " d" + ref
            probe_id = "%s_%d" % (probe, i % n_probes_per_subject)
            if to_csv:
                csv_writer.writerow(
                    [ref, probe, probe_id, score] + list(metadata.values())
                )
            else:
                f.write(
                    "%s%s%s %s %f\n" % (ref, s_five, probe, probe_id, score)
                )

        logger.debug("Writing unknown scores.")
        if neg_unknown is not None:
            s_unknown_subjects = ["u%d" % i for i in range(n_unknown_subjects)]
            for i, score in enumerate(neg_unknown):
                ref = s_subjects[
                    int(i / n_probes_per_subject / n_unknown_subjects)
                    % n_subjects
                ]
                probe = s_unknown_subjects[
                    int(i / n_probes_per_subject) % n_unknown_subjects
                ]
                s_five = " " if not five_col else " d" + ref + " "
                probe_id = "%s_%d" % (probe, i % n_probes_per_subject)
                if to_csv:
                    csv_writer.writerow(
                        [ref, probe, probe_id, score] + list(metadata.values())
                    )
                else:
                    f.write(
                        "%s%s%s %s %f\n" % (ref, s_five, probe, probe_id, score)
                    )


@click.command(
    epilog="""
Scores generation examples:

Output 'scores-dev.csv' and 'scores-eval.csv' in a new folder 'generated_scores/':

  $ bob bio gen ./generated_scores

Output scores similar to a system evaluated on the AT&T dataset dev group:

  $ bob bio gen -s 20 -p 5 ./generated_scores

Output a given number of scores in each file:

  $ bob bio gen -f --n-neg 500 --n-pos 100 ./generated_scores

Include unknown subjects scores:

  $ bob bio gen -s 5 -u 2 ./generated_scores

Change the mean and standard deviation of the scores distributions:

  $ bob bio gen -mm 1 -sp 0.3 -mnm -1 -sn 0.5 ./generated_scores

You can observe the distributions histograms in a pdf file with:

  $ bob bio hist -e ./generated_scores/scores-{dev,eval}.csv -o hist_gen.pdf
"""
)
@click.argument("outdir")
@click.option(
    "-mm",
    "--mean-match",
    default=10,
    type=click.FLOAT,
    show_default=True,
    help="Mean for the positive scores distribution",
)
@click.option(
    "-mnm",
    "--mean-non-match",
    default=-10,
    type=click.FLOAT,
    show_default=True,
    help="Mean for the negative scores distribution",
)
@click.option(
    "-p",
    "--n-probes-per-subject",
    default=5,
    type=click.INT,
    show_default=True,
    help="Number of probes per subject",
)
@click.option(
    "-s",
    "--n-subjects",
    default=50,
    type=click.INT,
    show_default=True,
    help="Number of subjects",
)
@click.option(
    "-sp",
    "--sigma-positive",
    default=10,
    type=click.FLOAT,
    show_default=True,
    help="Variance for the positive score distributions",
)
@click.option(
    "-sn",
    "--sigma-negative",
    default=10,
    type=click.FLOAT,
    show_default=True,
    help="Variance for the negative score distributions",
)
@click.option(
    "-u",
    "--n-unknown-subjects",
    default=0,
    type=click.INT,
    show_default=True,
    help="Number of unknown subjects (useful for open-set plots)",
)
@click.option(
    "-f",
    "--force-count",
    "force_count",
    is_flag=True,
    help="Use --n-pos and --n-neg amounts instead of the subject and sample counts",
)
@click.option(
    "--n-pos",
    "n_pos",
    default=5000,
    type=click.INT,
    show_default=True,
    help="Number of Positive verifications (number of lines in the file)",
)
@click.option(
    "--n-neg",
    "n_neg",
    default=5000,
    type=click.INT,
    show_default=True,
    help="Number of Negative verifications (number of lines in the file)",
)
@click.option(
    "--n-unk",
    "n_unk",
    default=5000,
    type=click.INT,
    show_default=True,
    help="Number of Unknown verifications (number of lines in the file)",
)
@click.option("--csv/--legacy", default=True, show_default=True)
@click.option("--five-col/--four-col", default=False, show_default=True)
@verbosity_option(logger=logger)
def gen(
    outdir,
    mean_match,
    mean_non_match,
    n_probes_per_subject,
    n_subjects,
    sigma_positive,
    sigma_negative,
    n_unknown_subjects,
    csv,
    five_col,
    force_count,
    n_pos,
    n_neg,
    n_unk,
    **kwargs,
):
    """Generate random scores.

    Generates random scores in 4col or 5col format. The scores are generated
    using Gaussian distribution whose mean and variance are an input
    parameter. The generated scores can be used as hypothetical datasets.

    This command generates scores relative to the number of subjects and
    probes per subjects, unless the -f flag is set. In that case, the --n-pos
    and --n-neg options are used as number of genuine and impostor
    comparisons.
    """

    # Compute the number of verifications needed
    if force_count:
        neg_count, pos_count, unknown_count = n_neg, n_pos, n_unk
    else:
        # One reference (model), and `n_probes_per_subject` probes per subject
        neg_count = n_subjects * n_probes_per_subject * (n_subjects - 1)
        pos_count = n_probes_per_subject * n_subjects
        unknown_count = n_unknown_subjects * n_subjects * n_probes_per_subject

    # Generate the data
    logger.info("Generating dev scores.")
    neg_dev, pos_dev = gen_score_distr(
        mean_non_match,
        mean_match,
        sigma_negative,
        sigma_positive,
        n_neg=neg_count,
        n_pos=pos_count,
        seed=0,
    )
    logger.info("Generating eval scores.")
    neg_eval, pos_eval = gen_score_distr(
        mean_non_match,
        mean_match,
        sigma_negative,
        sigma_positive,
        n_neg=neg_count,
        n_pos=pos_count,
        seed=1,
    )

    # For simplicity I will use the same distribution for dev-eval
    if n_unknown_subjects:
        logger.info("Generating unknown scores.")
        neg_unknown, _ = gen_score_distr(
            mean_non_match,
            mean_match,
            sigma_negative,
            sigma_positive,
            n_neg=unknown_count,
            n_pos=0,
            seed=2,
        )
    else:
        neg_unknown = None

    # Write the data into files
    logger.info("Saving results.")
    write_scores_to_file(
        neg_dev,
        pos_dev,
        os.path.join(outdir, "scores-dev.csv"),
        n_subjects,
        n_probes_per_subject,
        n_unknown_subjects,
        neg_unknown,
        csv,
        five_col,
    )

    write_scores_to_file(
        neg_eval,
        pos_eval,
        os.path.join(outdir, "scores-eval.csv"),
        n_subjects,
        n_probes_per_subject,
        n_unknown_subjects,
        neg_unknown,
        csv,
        five_col,
    )



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

import abc
import os

from .file import BioFile
from .legacy import FileDatabase as LegacyFileDatabase


class BioDatabase(LegacyFileDatabase, metaclass=abc.ABCMeta):
    """This class represents the basic API for database access.
    Please use this class as a base class for your database access classes.
    Do not forget to call the constructor of this base class in your derived class.

    **Parameters:**

    name : str
    A unique name for the database.

    all_files_options : dict
    Dictionary of options passed to the :py:meth:`bob.bio.base.database.BioDatabase.objects` database query when retrieving all data.

    extractor_training_options : dict
    Dictionary of options passed to the :py:meth:`bob.bio.base.database.BioDatabase.objects` database query used to retrieve the files for the extractor training.

    projector_training_options : dict
    Dictionary of options passed to the :py:meth:`bob.bio.base.database.BioDatabase.objects` database query used to retrieve the files for the projector training.

    enroller_training_options : dict
    Dictionary of options passed to the :py:meth:`bob.bio.base.database.BioDatabase.objects` database query used to retrieve the files for the enroller training.

    check_original_files_for_existence : bool
    Enables to test for the original data files when querying the database.

    original_directory : str
    The directory where the original data of the database are stored.

    original_extension : str
    The file name extension of the original data.

    annotation_directory : str
    The directory where the image annotations of the database are stored, if any.

    annotation_extension : str
    The file name extension of the annotation files.

    annotation_type : str
    The type of the annotation file to read, only json works.

    protocol : str or ``None``
    The name of the protocol that defines the default experimental setup for this database.

    training_depends_on_protocol : bool
    Specifies, if the training set used for training the extractor and the projector depend on the protocol.
    This flag is used to avoid re-computation of data when running on the different protocols of the same database.

    models_depend_on_protocol : bool
    Specifies, if the models depend on the protocol.
    This flag is used to avoid re-computation of models when running on the different protocols of the same database.

    kwargs : ``key=value`` pairs
    The arguments of the `Database` base class constructor.

    """

    # tell test runners (such as nose and pytest) that this class is not a test class
    ___test___ = False

    def __init__(
        self,
        name,
        all_files_options={},  # additional options for the database query that can be used to extract all files
        extractor_training_options={},
        # additional options for the database query that can be used to extract the training files for the extractor training
        projector_training_options={},
        # additional options for the database query that can be used to extract the training files for the extractor training
        enroller_training_options={},
        # additional options for the database query that can be used to extract the training files for the extractor training
        check_original_files_for_existence=False,
        original_directory=None,
        original_extension=None,
        annotation_directory=None,
        annotation_extension=None,
        annotation_type=None,
        protocol="Default",
        training_depends_on_protocol=False,
        models_depend_on_protocol=False,
        **kwargs,
    ):
        assert isinstance(name, str)

        super(BioDatabase, self).__init__(
            original_directory=original_directory,
            original_extension=original_extension,
            **kwargs,
        )

        self.name = name

        self.all_files_options = all_files_options
        self.extractor_training_options = extractor_training_options
        self.projector_training_options = projector_training_options
        self.enroller_training_options = enroller_training_options
        self.check_existence = check_original_files_for_existence

        self._kwargs = {}

        self.annotation_directory = annotation_directory
        self.annotation_extension = annotation_extension or ".json"
        self.annotation_type = annotation_type or "json"
        self.protocol = protocol
        self.training_depends_on_protocol = training_depends_on_protocol
        self.models_depend_on_protocol = models_depend_on_protocol
        self.models_depend_on_protocol = models_depend_on_protocol

        # try if the implemented model_ids_with_protocol() and objects() function have at least the required interface
        try:
            # create a value that is very unlikely a valid value for anything
            test_value = "#6T7+§X"
            # test if the parameters of the functions apply
            self.model_ids_with_protocol(groups=test_value, protocol=test_value)
            self.objects(
                groups=test_value,
                protocol=test_value,
                purposes=test_value,
                model_ids=(test_value,),
            )
            self.annotations(file=BioFile(test_value, test_value, test_value))
        except TypeError as e:
            # type error indicates that the given parameters are not valid.
            raise NotImplementedError(
                str(e)
                + "\nPlease implement:\n - the model_ids_with_protocol(...) function with at least the "
                "arguments 'groups' and 'protocol'\n - the objects(...) function with at least the "
                "arguments 'groups', 'protocol', 'purposes' and 'model_ids'\n - the annotations() "
                "function with at least the arguments 'file_id'."
            )
        except Exception:
            # any other error is fine at this stage.
            pass

    def __str__(self):
        """__str__() -> info

        This function returns all parameters of this class.

        **Returns:**

        info : str
          A string containing the full information of all parameters of this class.
        """
        params = (
            "name=%s, protocol=%s, original_directory=%s, original_extension=%s"
            % (
                self.name,
                self.protocol,
                self.original_directory,
                self.original_extension,
            )
        )
        params += ", ".join(
            ["%s=%s" % (key, value) for key, value in self._kwargs.items()]
        )
        params += ", original_directory=%s, original_extension=%s" % (
            self.original_directory,
            self.original_extension,
        )
        if self.all_files_options:
            params += ", all_files_options=%s" % self.all_files_options
        if self.extractor_training_options:
            params += (
                ", extractor_training_options=%s"
                % self.extractor_training_options
            )
        if self.projector_training_options:
            params += (
                ", projector_training_options=%s"
                % self.projector_training_options
            )
        if self.enroller_training_options:
            params += (
                ", enroller_training_options=%s"
                % self.enroller_training_options
            )

        return "%s(%s)" % (str(self.__class__), params)

    def replace_directories(self, replacements=None):
        """This helper function replaces the ``original_directory`` and the ``annotation_directory`` of the database with the directories read from the given replacement file.

        This function is provided for convenience, so that the database configuration files do not need to be modified.
        Instead, this function uses the given dictionary of replacements to change the original directory and the original extension (if given).

        The given ``replacements`` can be of type ``dict``, including all replacements, or a file name (as a ``str``), in which case the file is read.
        The structure of the file should be:

        .. code-block:: text

           # Comments starting with # and empty lines are ignored

           [YOUR_..._DATA_DIRECTORY] = /path/to/your/data
           [YOUR_..._ANNOTATION_DIRECTORY] = /path/to/your/annotations

        If no annotation files are available (e.g. when they are stored inside the ``database``), the annotation directory can be left out.

        **Parameters:**

        replacements : dict or str
          A dictionary with replacements, or a name of a file to read the dictionary from.
          If the file name does not exist, no directories are replaced.
        """
        if replacements is None:
            return
        if isinstance(replacements, str):
            if not os.path.exists(replacements):
                return
            # Open the database replacement file and reads its content
            with open(replacements) as f:
                replacements = {}
                for line in f:
                    if line.strip() and not line.startswith("#"):
                        splits = line.split("=")
                        assert len(splits) == 2
                        replacements[splits[0].strip()] = splits[1].strip()

        assert isinstance(replacements, dict)

        if self.original_directory in replacements:
            self.original_directory = replacements[self.original_directory]

        try:
            if self.annotation_directory in replacements:
                self.annotation_directory = replacements[
                    self.annotation_directory
                ]
        except AttributeError:
            pass

    ###########################################################################
    # Helper functions that you might want to use in derived classes
    ###########################################################################
    def uses_probe_file_sets(self, protocol=None):
        """Defines if, for the current protocol, the database uses several probe files to generate a score.
        Returns True if the given protocol specifies file sets for probes, instead of a single probe file.
        In this default implementation, False is returned, throughout.
        If you need different behavior, please overload this function in your derived class.
        """
        return False

    def arrange_by_client(self, files):
        """arrange_by_client(files) -> files_by_client

        Arranges the given list of files by client id.
        This function returns a list of lists of File's.

        **Parameters:**

        files : :py:class:`bob.bio.base.database.BioFile`
          A list of files that should be split up by `BioFile.client_id`.

        **Returns:**

        files_by_client : [[:py:class:`bob.bio.base.database.BioFile`]]
          The list of lists of files, where each sub-list groups the files with the same `BioFile.client_id`
        """
        client_files = {}
        for file in files:
            if file.client_id not in client_files:
                client_files[file.client_id] = []
            client_files[file.client_id].append(file)

        files_by_clients = []
        for client in sorted(client_files.keys()):
            files_by_clients.append(client_files[client])
        return files_by_clients

    def file_names(self, files, directory, extension):
        """file_names(files, directory, extension) -> paths

        Returns the full path of the given File objects.

        **Parameters:**

        files : [:py:class:`bob.bio.base.database.BioFile`]
          The list of file object to retrieve the file names for.

        directory : str
          The base directory, where the files can be found.

        extension : str
          The file name extension to add to all files.

        **Returns:**

        paths : [str] or [[str]]
          The paths extracted for the files, in the same order.
          If this database provides file sets, a list of lists of file names is returned, one sub-list for each file set.
        """
        # return the paths of the files
        if self.uses_probe_file_sets() and files and hasattr(files[0], "files"):
            # List of Filesets: do not remove duplicates
            return [
                [f.make_path(directory, extension) for f in file_set.files]
                for file_set in files
            ]
        else:
            # List of files, do not remove duplicates
            return [f.make_path(directory, extension) for f in files]

    #################################################################
    # Methods to be overwritten by derived classes
    #################################################################
    @abc.abstractmethod
    def model_ids_with_protocol(self, groups=None, protocol=None, **kwargs):
        """model_ids_with_protocol(groups = None, protocol = None, **kwargs) -> ids

        Returns a list of model ids for the given groups and given protocol.

        **Parameters:**

        groups : one or more of ``('world', 'dev', 'eval')``
          The groups to get the model ids for.

        protocol: a protocol name

        **Returns:**

        ids : [int] or [str]
          The list of (unique) model ids for the given groups.
        """
        raise NotImplementedError(
            "Please implement this function in derived classes"
        )

    def groups(self, protocol=None):
        """
        Returns the names of all registered groups in the database

        Keyword parameters:

        protocol: str
          The protocol for which the groups should be retrieved.
          If you do not have protocols defined, just ignore this field.
        """
        raise NotImplementedError(
            "This function must be implemented in your derived class."
        )

    @abc.abstractmethod
    def objects(
        self,
        groups=None,
        protocol=None,
        purposes=None,
        model_ids=None,
        **kwargs,
    ):
        """This function returns a list of :py:class:`bob.bio.base.database.BioFile` objects or the list
        of objects which inherit from this class. Returned files fulfill the given restrictions.

        Keyword parameters:

        groups : str or [str]
          The groups of which the clients should be returned.
          Usually, groups are one or more elements of ('world', 'dev', 'eval')

        protocol
          The protocol for which the clients should be retrieved.
          The protocol is dependent on your database.
          If you do not have protocols defined, just ignore this field.

        purposes : str or [str]
          The purposes for which File objects should be retrieved.
          Usually, purposes are one of ('enroll', 'probe').

        model_ids : [various type]
          The model ids for which the File objects should be retrieved.
          What defines a 'model id' is dependent on the database.
          In cases, where there is only one model per client, model ids and client ids are identical.
          In cases, where there is one model per file, model ids and file ids are identical.
          But, there might also be other cases.
        """
        raise NotImplementedError(
            "This function must be implemented in your derived class."
        )

    def annotations(self, file):
        """
        Returns the annotations for the given File object, if available.
        You need to override this method in your high-level implementation.
        If your database does not have annotations, it should return ``None``.

        **Parameters:**

        file : :py:class:`bob.bio.base.database.BioFile`
          The file for which annotations should be returned.

        **Returns:**

        annots : dict or None
          The annotations for the file, if available.
        """
        raise NotImplementedError(
            "This function must be implemented in your derived class."
        )

    #################################################################
    # Methods to provide common functionality
    #################################################################

    def model_ids(self, groups="dev"):
        """model_ids(group = 'dev') -> ids

        Returns a list of model ids for the given group, respecting the current protocol.

        **Parameters:**

        group : one of ``('dev', 'eval')``
          The group to get the model ids for.

        **Returns:**

        ids : [int] or [str]
          The list of (unique) model ids for models of the given group.
        """
        return sorted(
            self.model_ids_with_protocol(groups=groups, protocol=self.protocol)
        )

    def all_files(self, groups=None, **kwargs):
        """all_files(groups=None) -> files

        Returns all files of the database, respecting the current protocol.
        The files can be limited using the ``all_files_options`` in the constructor.

        **Parameters:**

        groups : some of ``('world', 'dev', 'eval')`` or ``None``
          The groups to get the data for.
          If ``None``, data for all groups is returned.

        kwargs: ignored

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFile`]
          The sorted and unique list of all files of the database.
        """
        return self.sort(
            self.objects(
                protocol=self.protocol, groups=groups, **self.all_files_options
            )
        )

    def training_files(self, step=None, arrange_by_client=False):
        """training_files(step = None, arrange_by_client = False) -> files

        Returns all training files for the given step, and arranges them by client, if desired, respecting the current protocol.
        The files for the steps can be limited using the ``..._training_options`` defined in the constructor.

        **Parameters:**

        step : one of ``('train_extractor', 'train_projector', 'train_enroller')`` or ``None``
          The step for which the training data should be returned.

        arrange_by_client : bool
          Should the training files be arranged by client?
          If set to ``True``, training files will be returned in [[:py:class:`bob.bio.base.database.BioFile`]], where each sub-list contains the files of a single client.
          Otherwise, all files will be stored in a simple [:py:class:`bob.bio.base.database.BioFile`].

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFile`] or [[:py:class:`bob.bio.base.database.BioFile`]]
          The (arranged) list of files used for the training of the given step.
        """
        if step is None:
            training_options = self.all_files_options
        elif step == "train_extractor":
            training_options = self.extractor_training_options
        elif step == "train_projector":
            training_options = self.projector_training_options
        elif step == "train_enroller":
            training_options = self.enroller_training_options
        else:
            raise ValueError(
                "The given step '%s' must be one of ('train_extractor', 'train_projector', 'train_enroller')"
                % step
            )

        files = self.sort(
            self.objects(
                protocol=self.protocol, groups="world", **training_options
            )
        )
        if arrange_by_client:
            return self.arrange_by_client(files)
        else:
            return files

    def test_files(self, groups=["dev"]):
        """test_files(groups = ['dev']) -> files

        Returns all test files (i.e., files used for enrollment and probing) for the given groups, respecting the current protocol.
        The files for the steps can be limited using the ``all_files_options`` defined in the constructor.

        **Parameters:**

        groups : some of ``('dev', 'eval')``
          The groups to get the data for.

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFile`]
          The sorted and unique list of test files of the database.
        """
        return self.sort(
            self.objects(
                protocol=self.protocol, groups=groups, **self.all_files_options
            )
        )

    def enroll_files(self, model_id=None, group="dev"):
        """enroll_files(model_id, group = 'dev') -> files

        Returns a list of File objects that should be used to enroll the model with the given model id from the given group, respecting the current protocol.
        If the model_id is None (the default), enrollment files for all models are returned.

        **Parameters:**

        model_id : int or str
          A unique ID that identifies the model.

        group : one of ``('dev', 'eval')``
          The group to get the enrollment files for.

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFile`]
          The list of files used for to enroll the model with the given model id.
        """
        if model_id:
            return self.sort(
                self.objects(
                    protocol=self.protocol,
                    groups=group,
                    model_ids=(model_id,),
                    purposes="enroll",
                    **self.all_files_options,
                )
            )
        else:
            return self.sort(
                self.objects(
                    protocol=self.protocol,
                    groups=group,
                    purposes="enroll",
                    **self.all_files_options,
                )
            )

    def probe_files(self, model_id=None, group="dev"):
        """probe_files(model_id = None, group = 'dev') -> files

        Returns a list of probe File objects, respecting the current protocol.
        If a ``model_id`` is specified, only the probe files that should be compared with the given model id are returned (for most databases, these are all probe files of the given group).
        Otherwise, all probe files of the given group are returned.

        **Parameters:**

        model_id : int or str or ``None``
          A unique ID that identifies the model.

        group : one of ``('dev', 'eval')``
          The group to get the enrollment files for.

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFile`]
          The list of files used for to probe the model with the given model id.
        """
        if model_id is not None:
            files = self.objects(
                protocol=self.protocol,
                groups=group,
                model_ids=(model_id,),
                purposes="probe",
                **self.all_files_options,
            )
        else:
            files = self.objects(
                protocol=self.protocol,
                groups=group,
                purposes="probe",
                **self.all_files_options,
            )
        return self.sort(files)

    def object_sets(
        self,
        groups=None,
        protocol=None,
        purposes=None,
        model_ids=None,
        **kwargs,
    ):
        """This function returns lists of FileSet objects, which fulfill the given restrictions.

        Keyword parameters:

        groups : str or [str]
          The groups of which the clients should be returned.
          Usually, groups are one or more elements of ('world', 'dev', 'eval')

        protocol
          The protocol for which the clients should be retrieved.
          The protocol is dependent on your database.
          If you do not have protocols defined, just ignore this field.

        purposes : str or [str]
          The purposes for which File objects should be retrieved.
          Usually, purposes are one of ('enroll', 'probe').

        model_ids : [various type]
          The model ids for which the File objects should be retrieved.
          What defines a 'model id' is dependent on the database.
          In cases, where there is only one model per client, model ids and client ids are identical.
          In cases, where there is one model per file, model ids and file ids are identical.
          But, there might also be other cases.
        """
        raise NotImplementedError(
            "This function must be implemented in your derived class."
        )

    def probe_file_sets(self, model_id=None, group="dev"):
        """probe_file_sets(model_id = None, group = 'dev') -> files

        Returns a list of probe FileSet objects, respecting the current protocol.
        If a ``model_id`` is specified, only the probe files that should be compared with the given model id are returned (for most databases, these are all probe files of the given group).
        Otherwise, all probe files of the given group are returned.

        **Parameters:**

        model_id : int or str or ``None``
          A unique ID that identifies the model.

        group : one of ``('dev', 'eval')``
          The group to get the enrollment files for.

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFileSet`] or something similar
          The list of file sets used to probe the model with the given model id.
        """
        if model_id is not None:
            file_sets = self.object_sets(
                protocol=self.protocol,
                groups=group,
                model_ids=(model_id,),
                purposes="probe",
                **self.all_files_options,
            )
        else:
            file_sets = self.object_sets(
                protocol=self.protocol,
                groups=group,
                purposes="probe",
                **self.all_files_options,
            )
        return self.sort(file_sets)

    def client_id_from_model_id(self, model_id, group="dev"):
        """Return the client id associated with the given model id.
        In this base class implementation, it is assumed that only one model is enrolled for each client and, thus, client id and model id are identical.
        All key word arguments are ignored.
        Please override this function in derived class implementations to change this behavior.
        """
        return model_id


class ZTBioDatabase(BioDatabase):
    """This class defines another set of abstract functions that need to be implemented if your database provides the interface for computing scores used for ZT-normalization."""

    def __init__(
        self, name, z_probe_options={}, **kwargs
    ):  # Limit the z-probes
        """**Construtctor Documentation**

        This constructor tests if all implemented functions take the correct arguments.
        All keyword parameters will be passed unaltered to the :py:class:`bob.bio.base.database.BioDatabase` constructor.
        """
        # call base class constructor
        super(ZTBioDatabase, self).__init__(name, **kwargs)

        self.z_probe_options = z_probe_options

        # try if the implemented tmodel_ids_with_protocol(), tobjects() and zobjects() function have at least the required interface
        try:
            # create a value that is very unlikely a valid value for anything
            test_value = "#F9S%3*Y"
            # test if the parameters of the functions apply
            self.tmodel_ids_with_protocol(
                groups=test_value, protocol=test_value
            )
            self.tobjects(
                groups=test_value, protocol=test_value, model_ids=test_value
            )
            self.zobjects(groups=test_value, protocol=test_value)
        except TypeError as e:
            # type error indicates that the given parameters are not valid.
            raise NotImplementedError(
                str(e)
                + "\nPlease implement:\n - the tmodel_ids_with_protocol(...) function with at least the "
                "arguments 'groups' and 'protocol'\n - the tobjects(...) function with at least the arguments "
                "'groups', 'protocol' and 'model_ids'\n - the zobjects(...) function with at "
                "least the arguments 'groups' and 'protocol'"
            )
        except Exception:
            # any other error is fine at this stage.
            pass

    @abc.abstractmethod
    def tobjects(self, groups=None, protocol=None, model_ids=None, **kwargs):
        """This function returns the File objects of the T-Norm models of the given groups for the given protocol and the given model ids.

        Keyword parameters:

        groups : str or [str]
          The groups of which the model ids should be returned.
          Usually, groups are one or more elements of ('dev', 'eval')

        protocol : str
          The protocol for which the model ids should be retrieved.
          The protocol is dependent on your database.
          If you do not have protocols defined, just ignore this field.

        model_ids : [various type]
          The model ids for which the File objects should be retrieved.
          What defines a 'model id' is dependent on the database.
          In cases, where there is only one model per client, model ids and client ids are identical.
          In cases, where there is one model per file, model ids and file ids are identical.
          But, there might also be other cases.
        """
        raise NotImplementedError(
            "This function must be implemented in your derived class."
        )

    @abc.abstractmethod
    def zobjects(self, groups=None, protocol=None, **kwargs):
        """This function returns the File objects of the Z-Norm impostor files of the given groups for the given protocol.

        Keyword parameters:

        groups : str or [str]
          The groups of which the model ids should be returned.
          Usually, groups are one or more elements of ('dev', 'eval')

        protocol : str
          The protocol for which the model ids should be retrieved.
          The protocol is dependent on your database.
          If you do not have protocols defined, just ignore this field.
        """
        raise NotImplementedError(
            "This function must be implemented in your derived class."
        )

    def all_files(self, groups=["dev"], add_zt_files=True):
        """all_files(groups=None) -> files

        Returns all files of the database, including those for ZT norm, respecting the current protocol.
        The files can be limited using the ``all_files_options`` and the the ``z_probe_options`` in the constructor.

        **Parameters:**

        groups : some of ``('world', 'dev', 'eval')`` or ``None``
          The groups to get the data for.
          If ``None``, data for all groups is returned.

        add_zt_files: bool
          If set (the default), files for ZT score normalization are added.

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFile`]
          The sorted and unique list of all files of the database.
        """
        files = self.objects(
            protocol=self.protocol, groups=groups, **self.all_files_options
        )

        # add all files that belong to the ZT-norm
        if add_zt_files and groups:
            for group in groups:
                if group == "world":
                    continue
                files += self.tobjects(
                    protocol=self.protocol, groups=group, model_ids=None
                )
                files += self.zobjects(
                    protocol=self.protocol, groups=group, **self.z_probe_options
                )
        elif add_zt_files:
            files += self.tobjects(
                protocol=self.protocol, groups=groups, model_ids=None
            )
            files += self.zobjects(
                protocol=self.protocol, groups=groups, **self.z_probe_options
            )
        return self.sort(files)

    @abc.abstractmethod
    def tmodel_ids_with_protocol(self, protocol=None, groups=None, **kwargs):
        """This function returns the ids of the T-Norm models of the given groups for the given protocol.

        Keyword parameters:

        groups : str or [str]
          The groups of which the model ids should be returned.
          Usually, groups are one or more elements of ('dev', 'eval')

        protocol : str
          The protocol for which the model ids should be retrieved.
          The protocol is dependent on your database.
          If you do not have protocols defined, just ignore this field.
        """
        raise NotImplementedError(
            "This function must be implemented in your derived class."
        )

    def t_model_ids(self, groups="dev"):
        """t_model_ids(group = 'dev') -> ids

        Returns a list of model ids of T-Norm models for the given group, respecting the current protocol.

        **Parameters:**

        group : one of ``('dev', 'eval')``
          The group to get the model ids for.

        **Returns:**

        ids : [int] or [str]
          The list of (unique) model ids for T-Norm models of the given group.
        """
        return sorted(
            self.tmodel_ids_with_protocol(protocol=self.protocol, groups=groups)
        )

    def t_enroll_files(self, t_model_id, group="dev"):
        """t_enroll_files(t_model_id, group = 'dev') -> files

        Returns a list of File objects that should be used to enroll the T-Norm model with the given model id from the given group, respecting the current protocol.

        **Parameters:**

        t_model_id : int or str
          A unique ID that identifies the model.

        group : one of ``('dev', 'eval')``
          The group to get the enrollment files for.

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFile`]
          The sorted list of files used for to enroll the model with the given model id.
        """
        return self.sort(
            self.tobjects(
                protocol=self.protocol, groups=group, model_ids=(t_model_id,)
            )
        )

    def z_probe_files(self, group="dev"):
        """z_probe_files(group = 'dev') -> files

        Returns a list of probe files used to compute the Z-Norm, respecting the current protocol.
        The Z-probe files can be limited using the ``z_probe_options`` in the query to :py:meth:`bob.bio.base.database.ZTBioDatabase.z_probe_files`

        **Parameters:**

        group : one of ``('dev', 'eval')``
          The group to get the Z-norm probe files for.

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFile`]
          The unique list of files used to compute the Z-norm.
        """
        return self.sort(
            self.zobjects(
                protocol=self.protocol, groups=group, **self.z_probe_options
            )
        )

    def z_probe_file_sets(self, group="dev"):
        """z_probe_file_sets(group = 'dev') -> files

        Returns a list of probe FileSet objects used to compute the Z-Norm.
        This function needs to be implemented in derived class implementations.

        **Parameters:**

        group : one of ``('dev', 'eval')``
          The group to get the Z-norm probe files for.

        **Returns:**

        files : [:py:class:`bob.bio.base.database.BioFileSet`]
          The unique list of file sets used to compute the Z-norm.
        """
        raise NotImplementedError(
            "Please implement this function in derived classes"
        )

    def client_id_from_t_model_id(self, t_model_id, group="dev"):
        """client_id_from_t_model_id(t_model_id, group = 'dev') -> client_id
        Returns the client id for the given T-Norm model id.
        In this base class implementation, we just use the :py:meth:`BioDatabase.client_id_from_model_id` function.
        Overload this function if you need another behavior.

        **Parameters:**

        t_model_id : int or str
          A unique ID that identifies the T-Norm model.
        group : one of ``('dev', 'eval')``
          The group to get the client ids for.

        **Returns:**

        client_id : [int] or [str]
          A unique ID that identifies the client, to which the T-Norm model belongs.
        """
        return self.client_id_from_model_id(t_model_id, group)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# @author: Manuel Guenther <Manuel.Guenther@idiap.ch>
# @date: Tue Oct  2 12:12:39 CEST 2012

import warnings

from .. import utils


class Preprocessor(object):
    """This is the base class for all preprocessors.
    It defines the minimum requirements for all derived proprocessor classes.

    **Parameters:**

    writes_data : bool
      Select, if the preprocessor actually writes preprocessed images, or if it is simply returning values.

    read_original_data: callable or ``None``
      This function is used to read the original data from file.
      It takes three inputs: A :py:class:`bob.bio.base.database.BioFile` (or one of its derivatives), the original directory (as ``str``) and the original extension (as ``str``).
      If ``None``, the default function :py:func:`bob.bio.base.read_original_data` is used.

    min_preprocessed_file_size: int
      The minimum file size of a saved preprocessd data in bytes. If the saved
      preprocessed data file size is smaller than this, it is assumed to be a
      corrupt file and the data will be processed again.

    kwargs : ``key=value`` pairs
      A list of keyword arguments to be written in the `__str__` function.
    """

    def __init__(
        self,
        writes_data=True,
        read_original_data=None,
        min_preprocessed_file_size=1000,
        **kwargs,
    ):
        # Each class needs to have a constructor taking
        # all the parameters that are required for the preprocessing as arguments
        self.writes_data = writes_data
        if read_original_data is None:
            read_original_data = utils.read_original_data
        self.read_original_data = read_original_data
        self.min_preprocessed_file_size = min_preprocessed_file_size
        self._kwargs = kwargs

        warnings.warn(
            "`bob.bio.base.preprocessor.Preprocessor` will be deprecated in 01/01/2021. "
            "Please, implement your biometric algorithm using `bob.pipelines` (https://gitlab.idiap.ch/bob/bob.pipelines).",
            DeprecationWarning,
        )

    # The call function (i.e. the operator() in C++ terms)
    def __call__(self, data, annotations):
        """__call__(data, annotations) -> data

        This is the call function that you have to overwrite in the derived class.
        The parameters that this function will receive are:

        **Parameters:**

        data : object
          The original data that needs preprocessing, usually a :py:class:`numpy.ndarray`, but might be different.

        annotations : {} or None
          The annotations (if any)  that belongs to the given ``data``; as a dictionary.
          The type of the annotation depends on your kind of problem.

        **Returns:**

        data : object
          The *preprocessed* data, usually a :py:class:`numpy.ndarray`, but might be different.
        """
        raise NotImplementedError(
            "Please overwrite this function in your derived class"
        )

    def __str__(self):
        """__str__() -> info

        This function returns all parameters of this class (and its derived class).

        **Returns:**

        info : str
          A string containing the full information of all parameters of this (and the derived) class.
        """
        return utils.pretty_print(self, self._kwargs)

    ############################################################
    # Special functions that might be overwritten on need
    ############################################################

    def write_data(self, data, data_file):
        """Writes the given *preprocessed* data to a file with the given name.
        In this base class implementation, we simply use :py:func:`bob.bio.base.save` for that.
        If you have a different format (e.g. not images), please overwrite this function.

        **Parameters:**

        data : object
          The preprocessed data, i.e., what is returned from `__call__`.

        data_file : str or :py:class:`h5py.File`
          The file open for writing, or the name of the file to write.
        """
        utils.save(data, data_file)

    def read_data(self, data_file):
        """read_data(data_file) -> data

        Reads the *preprocessed* data from file.
        In this base class implementation, it uses :py:func:`bob.bio.base.load` to do that.
        If you have different format, please overwrite this function.

        **Parameters:**

        data_file : str or :py:class:`h5py.File`
          The file open for reading or the name of the file to read from.

        **Returns:**

        data : object (usually :py:class:`numpy.ndarray`)
          The preprocessed data read from file.
        """
        return utils.load(data_file)



"""The main entry for bob.vuln
"""
import click
import pkg_resources

from clapper.click import AliasedGroup
from click_plugins import with_plugins


@with_plugins(pkg_resources.iter_entry_points("bob.vuln.cli"))
@click.group(cls=AliasedGroup)
def vulnerability():
    """Vulnerability analysis related commands."""
    pass



""" Click commands for ``bob.bio.base`` """
import logging

import click

from clapper.click import verbosity_option

import bob.measure.script.figure as measure_figure

from bob.measure.script import common_options

from ..score import load
from . import figure as bio_figure

logger = logging.getLogger(__name__)

SCORE_FORMAT = (
    "Files must be 4- or 5- columns format, see "
    ":py:func:`bob.bio.base.score.load.four_column` and "
    ":py:func:`bob.bio.base.score.load.five_column` for details."
)
CRITERIA = ("eer", "min-hter", "far", "mindcf", "cllr", "rr")


def rank_option(**kwargs):
    """Get option for rank parameter"""

    def custom_rank_option(func):
        def callback(ctx, param, value):
            value = 1 if value < 0 else value
            ctx.meta["rank"] = value
            return value

        return click.option(
            "-rk",
            "--rank",
            type=click.INT,
            default=1,
            help="Provide rank for the command",
            callback=callback,
            show_default=True,
            **kwargs,
        )(func)

    return custom_rank_option


@common_options.metrics_command(
    common_options.METRICS_HELP.format(
        names="FtA, FAR, FRR, FMR, FNMR, HTER",
        criteria=CRITERIA,
        score_format=SCORE_FORMAT,
        hter_note="Note that FAR = FMR * (1 - FtA), FRR = FtA + FNMR * (1 - FtA) "
        "and HTER = (FMR + FNMR) / 2",
        command="bob bio metrics",
    ),
    criteria=CRITERIA,
)
@common_options.cost_option()
def metrics(ctx, scores, evaluation, **kwargs):
    if "criterion" in ctx.meta and ctx.meta["criterion"] == "rr":
        process = bio_figure.Metrics(ctx, scores, evaluation, load.cmc)
    else:
        process = bio_figure.Metrics(ctx, scores, evaluation, load.split)
    process.run()


@common_options.roc_command(
    common_options.ROC_HELP.format(
        score_format=SCORE_FORMAT, command="bob bio roc"
    )
)
def roc(ctx, scores, evaluation, **kwargs):
    process = bio_figure.Roc(ctx, scores, evaluation, load.split)
    process.run()


@common_options.det_command(
    common_options.DET_HELP.format(
        score_format=SCORE_FORMAT, command="bob bio det"
    )
)
def det(ctx, scores, evaluation, **kwargs):
    process = bio_figure.Det(ctx, scores, evaluation, load.split)
    process.run()


@common_options.epc_command(
    common_options.EPC_HELP.format(
        score_format=SCORE_FORMAT, command="bob bio epc"
    )
)
def epc(ctx, scores, **kwargs):
    process = measure_figure.Epc(ctx, scores, True, load.split)
    process.run()


@common_options.hist_command(
    common_options.HIST_HELP.format(
        score_format=SCORE_FORMAT, command="bob bio hist"
    )
)
def hist(ctx, scores, evaluation, **kwargs):
    process = bio_figure.Hist(ctx, scores, evaluation, load.split)
    process.run()


@common_options.evaluate_command(
    common_options.EVALUATE_HELP.format(
        score_format=SCORE_FORMAT, command="bob bio evaluate"
    ),
    criteria=CRITERIA,
)
@common_options.cost_option()
def evaluate(ctx, scores, evaluation, **kwargs):
    common_options.evaluate_flow(
        ctx, scores, evaluation, metrics, roc, det, epc, hist, **kwargs
    )


@common_options.multi_metrics_command(
    common_options.MULTI_METRICS_HELP.format(
        names="FtA, FAR, FRR, FMR, FNMR, HTER",
        criteria=CRITERIA,
        score_format=SCORE_FORMAT,
        command="bob bio multi-metrics",
    ),
    criteria=CRITERIA,
)
def multi_metrics(ctx, scores, evaluation, protocols_number, **kwargs):
    ctx.meta["min_arg"] = protocols_number * (2 if evaluation else 1)
    process = bio_figure.MultiMetrics(ctx, scores, evaluation, load.split)
    process.run()


@click.command()
@common_options.scores_argument(nargs=-1)
@common_options.titles_option()
@common_options.legends_option()
@common_options.sep_dev_eval_option()
@common_options.output_plot_file_option(default_out="cmc.pdf")
@common_options.eval_option()
@common_options.semilogx_option(True)
@common_options.axes_val_option(dflt=None)
@common_options.x_rotation_option()
@common_options.const_layout_option()
@common_options.style_option()
@common_options.linestyles_option()
@common_options.figsize_option()
@verbosity_option(logger=logger)
@click.pass_context
def cmc(ctx, scores, evaluation, **kwargs):
    """Plot CMC (cumulative match characteristic curve).
    graphical presentation of results of an identification task eval, plotting
    rank values on the x-axis and the probability of correct identification at
    or below that rank on the y-axis. The values for the axis will be computed
    using :py:func:`bob.measure.cmc`.

    You need to provide one or more development score file(s) for each
    experiment. You can also provide eval files along with dev files. If
    eval-scores are used, the flag `--eval` must be used. is required
    in that case. Files must be 4- or 5- columns format, see
    :py:func:`bob.bio.base.score.load.four_column` and
    :py:func:`bob.bio.base.score.load.five_column` for details.


    Examples:
        $ bob bio cmc -v dev-scores

        $ bob bio cmc -v dev-scores1 eval-scores1 dev-scores2
        eval-scores2

        $ bob bio cmc -v -o my_roc.pdf dev-scores1 eval-scores1
    """
    process = bio_figure.Cmc(ctx, scores, evaluation, load.cmc)
    process.run()


@click.command()
@common_options.scores_argument(nargs=-1)
@common_options.titles_option()
@common_options.legends_option()
@common_options.sep_dev_eval_option()
@common_options.output_plot_file_option(default_out="dir.pdf")
@common_options.eval_option()
@common_options.semilogx_option(True)
@common_options.axes_val_option(dflt=None)
@common_options.x_rotation_option()
@rank_option()
@common_options.const_layout_option()
@common_options.style_option()
@common_options.linestyles_option()
@common_options.figsize_option()
@common_options.min_far_option()
@verbosity_option(logger=logger)
@click.pass_context
def dir(ctx, scores, evaluation, **kwargs):
    """Plots the Detection & Identification Rate curve over the FAR.

    This curve is designed to be used in an open set identification protocol,
    and defined in Chapter 14.1 of [LiJain2005]_.  It requires to have at least
    one open set probe item, i.e., with no corresponding gallery, such that the
    positives for that pair are ``None``.

    The detection and identification curve first computes FAR thresholds based
    on the out-of-set probe scores (negative scores).  For each probe item, the
    **maximum** negative score is used.  Then, it plots the detection and
    identification rates for those thresholds, which are based on the in-set
    probe scores only. See [LiJain2005]_ for more details.

    .. [LiJain2005] **Stan Li and Anil K. Jain**, *Handbook of Face Recognition*, Springer, 2005

    You need to provide one or more development score file(s) for each
    experiment. You can also provide eval files along with dev files. If
    eval-scores are used, the flag `--eval` must be used. is required
    in that case. Files must be 4- or 5- columns format, see
    :py:func:`bob.bio.base.score.load.four_column` and
    :py:func:`bob.bio.base.score.load.five_column` for details.

    Examples:
        $ bob bio dir -e -v dev-scores

        $ bob bio dir -v dev-scores1 eval-scores1 dev-scores2
        eval-scores2

        $ bob bio dir -v -o my_roc.pdf dev-scores1 eval-scores1
    """
    process = bio_figure.Dir(ctx, scores, evaluation, load.cmc)
    process.run()



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Mon 23 May 2011 16:23:05 CEST

"""A set of utilities to load score files with different formats.
"""

import csv
import logging
import os
import tarfile

from pathlib import Path

import dask.dataframe
import numpy

logger = logging.getLogger(__name__)


def iscsv(filename):
    return ".csv" in Path(filename).suffixes


def open_file(filename, mode="rt"):
    """Opens the given score file for reading.

    Score files might be raw text files, or a tar-file including a single score
    file inside.


    Parameters:

      filename (:py:class:`str`, ``file-like``): The name of the score file to
        open, or a file-like object open for reading. If a file name is given,
        the according file might be a raw text file or a (compressed) tar file
        containing a raw text file.


    Returns:


      ``file-like``: A read-only file-like object as it would be returned by
      :py:func:`open`.

    """

    if not isinstance(filename, str) and hasattr(filename, "read"):
        # It seems that this is an open file
        return filename

    if not os.path.isfile(filename):
        raise IOError("Score file '%s' does not exist." % filename)
    if not tarfile.is_tarfile(filename):
        return open(filename, mode)

    # open the tar file for reading
    tar = tarfile.open(filename, "r")
    # get the first file in the tar file
    tar_info = tar.next()
    while tar_info is not None and not tar_info.isfile():
        tar_info = tar.next()
    # check that one file was found in the archive
    if tar_info is None:
        raise IOError(
            "The given file is a .tar file, but it does not contain any file."
        )

    # open the file for reading
    return tar.extractfile(tar_info)


def four_column(filename):
    """Loads a score set from a single file and yield its lines

    Loads a score set from a single file and yield its lines (to avoid loading
    the score file at once into memory).  This function verifies that all fields
    are correctly placed and contain valid fields.  The score file must contain
    the following information in each line:

    .. code-block:: text

       claimed_id real_id test_label score


    Parameters:

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.


    Yields:

      str: The claimed identity -- the client name of the model that was used in
      the comparison

      str: The real identity -- the client name of the probe that was used in the
      comparison

      str: A label of the probe -- usually the probe file name, or the probe id

      float: The result of the comparison of the model and the probe

    """
    return _iterate_score_file(filename)


def split_four_column(filename):
    """Loads a score set from a single file and splits the scores

    Loads a score set from a single file and splits the scores between negatives
    and positives. The score file has to respect the 4 column format as defined
    in the method :py:func:`four_column`.

    This method avoids loading and allocating memory for the strings present in
    the file. We only keep the scores.


    Parameters:

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.


    Returns:

      array: negatives, 1D float array containing the list of scores, for which
        the ``claimed_id`` and the ``real_id`` are different
        (see :py:func:`four_column`)

      array: positives, 1D float array containing the list of scores, for which
        the ``claimed_id`` and the ``real_id`` are identical
        (see :py:func:`four_column`)

    """

    score_lines = four_column(filename)
    return _split_scores(score_lines, 1)


def get_split_dataframe(filename):
    """Loads a score set that was written with :any:`bob.bio.base.pipelines.CSVScoreWriter`

    Returns two dataframes, split between positives and negatives.

    Parameters
    ----------

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.

    Returns
    -------

      dataframe: negatives, contains the list of scores (and metadata) for which
        the fields of the ``bio_ref_subject_id`` and ``probe_subject_id`` columns are
        different. (see :ref:`bob.bio.base.pipeline_simple_advanced_features`)

      dataframe: positives, contains the list of scores (and metadata) for which
        the fields of the ``bio_ref_subject_id`` and ``probe_subject_id`` columns are
        identical. (see :ref:`bob.bio.base.pipeline_simple_advanced_features`)

    """
    df = dask.dataframe.read_csv(filename)

    genuines = df[df.probe_subject_id == df.bio_ref_subject_id]
    impostors = df[df.probe_subject_id != df.bio_ref_subject_id]

    return impostors, genuines


def split_csv_scores(filename):
    """Loads a score set that was written with :any:`bob.bio.base.pipelines.CSVScoreWriter`

    Parameters
    ----------

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.

    Returns
    -------

      array: negatives, 1D float array containing the list of scores, for which
        the fields of the ``bio_ref_subject_id`` and ``probe_subject_id`` columns are
        different. (see :ref:`bob.bio.base.pipeline_simple_advanced_features`)

      array: positives, 1D float array containing the list of scores, for which
        the fields of the ``bio_ref_subject_id`` and ``probe_subject_id`` columns are
        identical. (see :ref:`bob.bio.base.pipeline_simple_advanced_features`)

    """
    df = dask.dataframe.read_csv(filename)

    genuines = df[df.probe_subject_id == df.bio_ref_subject_id]
    impostors = df[df.probe_subject_id != df.bio_ref_subject_id]

    return (
        impostors["score"].to_dask_array().compute(),
        genuines["score"].to_dask_array().compute(),
    )


def cmc_four_column(filename):
    """Loads scores to compute CMC curves from a file in four column format.

    The four column file needs to be in the same format as described in
    :py:func:`four_column`, and the ``test_label`` (column 3) has to contain the
    test/probe file name or a probe id.

    This function returns a list of tuples.  For each probe file, the tuple
    consists of a list of negative scores and a list of positive scores.
    Usually, the list of positive scores should contain only one element, but
    more are allowed.  The result of this function can directly be passed to,
    e.g., the :py:func:`bob.measure.cmc` function.


    Parameters:

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.


    Returns:

      :any:`list`: A list of tuples, where each tuple contains the
      ``negative`` and ``positive`` scores for one probe of the database. Both
      ``negatives`` and ``positives`` can be either an 1D
      :py:class:`numpy.ndarray` of type ``float``, or ``None``.

    """

    score_lines = four_column(filename)
    return _split_cmc_scores(score_lines, 1)


def five_column(filename):
    """Loads a score set from a single file and yield its lines

    Loads a score set from a single file and yield its lines (to avoid loading
    the score file at once into memory).  This function verifies that all fields
    are correctly placed and contain valid fields.  The score file must contain
    the following information in each line:

    .. code-block:: text

       claimed_id model_label real_id test_label score


    Parameters:

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.


    Yields:

      str: The claimed identity -- the client name of the model that was used in
      the comparison

      str: A label for the model -- usually the model file name, or the model id

      str: The real identity -- the client name of the probe that was used in the
      comparison

      str: A label of the probe -- usually the probe file name, or the probe id

      float: The result of the comparison of the model and the probe

    """

    return _iterate_score_file(filename)


def split_five_column(filename):
    """Loads a score set from a single file and splits the scores

    Loads a score set from a single file in five column format and splits the
    scores between negatives and positives. The score file has to respect the 5
    column format as defined in the method :py:func:`five_column`.

    This method avoids loading and allocating memory for the strings present in
    the file. We only keep the scores.


    Parameters:

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.


    Returns:

      array: negatives, 1D float array containing the list of scores, for which
        the ``claimed_id`` and the ``real_id`` are different
        (see :py:func:`four_column`)

      array: positives, 1D float array containing the list of scores, for which
        the ``claimed_id`` and the ``real_id`` are identical
        (see :py:func:`four_column`)

    """

    score_lines = four_column(filename)
    return _split_scores(score_lines, 2)


def cmc_five_column(filename):
    """Loads scores to compute CMC curves from a file in five column format.

    The five column file needs to be in the same format as described in
    :py:func:`five_column`, and the ``test_label`` (column 4) has to contain the
    test/probe file name or a probe id.

    This function returns a list of tuples.  For each probe file, the tuple
    consists of a list of negative scores and a list of positive scores.
    Usually, the list of positive scores should contain only one element, but
    more are allowed.  The result of this function can directly be passed to,
    e.g., the :py:func:`bob.measure.cmc` function.


    Parameters:

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.


    Returns:

      :any:`list`: A list of tuples, where each tuple contains the
      ``negative`` and ``positive`` scores for one probe of the database.

    """
    score_lines = four_column(filename)
    return _split_cmc_scores(score_lines, 2)


def scores(filename, ncolumns=None):
    """Loads the scores from the given score file and yield its lines.
    Depending on the score file format, four or five elements are yielded, see
    :py:func:`bob.bio.base.score.load.four_column` and
    :py:func:`bob.bio.base.score.load.five_column` for details.

    Parameters:

    filename:  :py:class:`str`, ``file-like``:
      The file object that will be opened with :py:func:`open_file` containing the scores.

    ncolumns: any
      ignored

    Yields:

    tuple:
      see :py:func:`bob.bio.base.score.load.four_column` or
      :py:func:`bob.bio.base.score.load.five_column`
    """
    return _iterate_score_file(filename)


def split(filename, ncolumns=None, sort=False):
    """Loads the scores from the given score file and splits them into positives
    and negatives.
    Depending on the score file format, it calls see
    :py:func:`bob.bio.base.score.load.split_four_column` and
    :py:func:`bob.bio.base.score.load.split_five_column` for details.

    Parameters
    ----------

    filename : str
        The path to the score file.
    ncolumns : int or ``None``
        If specified to be ``4`` or ``5``, the score file will be assumed to be
        in the given format. If not specified, the score file format will be
        estimated automatically
    sort : :obj:`bool`, optional
        If ``True``, will return sorted negatives and positives

    Returns
    -------

    negatives : 1D :py:class:`numpy.ndarray` of type float
        This array contains the list of scores, for which the ``claimed_id`` and
        the ``real_id`` are different (see :py:func:`four_column`)
    positives : 1D :py:class:`numpy.ndarray` of type float
        This array contains the list of scores, for which the ``claimed_id`` and
        the ``real_id`` are identical (see :py:func:`four_column`)
    """
    if iscsv(filename):
        neg, pos = split_csv_scores(filename)
    else:
        ncolumns = _estimate_score_file_format(filename, ncolumns)
        if ncolumns == 4:
            neg, pos = split_four_column(filename)
        else:
            assert ncolumns == 5
            neg, pos = split_five_column(filename)

    if sort:
        neg.sort()
        pos.sort()

    return neg, pos


def cmc(filename, ncolumns=None):
    """cmc(filename, ncolumns=None) -> list

    Loads scores to compute CMC curves.

    Depending on the score file format, it calls see
    :py:func:`bob.bio.base.score.load.cmc_four_column` and
    `:py:func:`bob.bio.base.score.load.cmc_five_column` for details.

    Parameters:

      filename (:py:class:`str` or ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.

      ncolumns: (:py:class:`int`, Optional): If specified to be ``4`` or ``5``,
        the score file will be assumed to be in the given format.  If not
        specified, the score file format will be estimated automatically

    Returns:

    :any:`list`: [(neg,pos)] A list of tuples, where each tuple contains the
    ``negative`` and ``positive`` scores for one probe of the database.

    """

    ncolumns = (
        4
        if iscsv(filename)
        else _estimate_score_file_format(filename, ncolumns)
    )

    if ncolumns == 4:
        return cmc_four_column(filename)
    else:
        assert ncolumns == 5
        return cmc_five_column(filename)


def load_score(filename, ncolumns=None, minimal=False, **kwargs):
    """Load scores using numpy.loadtxt and return the data as a numpy array.

    Parameters:

      filename (:py:class:`str`, ``file-like``): The file object that will be
        opened with :py:func:`open_file` containing the scores.

      ncolumns (:py:class:`int`, optional): 4, 5 or None (the default),
        specifying the number of columns in the score file. If None is provided,
        the number of columns will be guessed.

      minimal (:py:class:`bool`, optional): If True, only loads ``claimed_id``, ``real_id``,
        and ``scores``.

      **kwargs: Keyword arguments passed to :py:func:`numpy.genfromtxt`


    Returns:

      array: An array which contains not only the actual scores but also the
      ``claimed_id``, ``real_id``, ``test_label`` and ``['model_label']``

    """

    def convertfunc(x):
        return x

    ncolumns = _estimate_score_file_format(filename, ncolumns)

    usecols = kwargs.pop("usecols", None)
    if ncolumns == 4:
        names = ("claimed_id", "real_id", "test_label", "score")
        converters = {0: convertfunc, 1: convertfunc, 2: convertfunc, 3: float}
        if minimal:
            usecols = (0, 1, 3)

    elif ncolumns == 5:
        names = ("claimed_id", "model_label", "real_id", "test_label", "score")
        converters = {
            0: convertfunc,
            1: convertfunc,
            2: convertfunc,
            3: convertfunc,
            4: float,
        }
        if minimal:
            usecols = (0, 2, 4)
    else:
        raise ValueError("ncolumns of 4 and 5 are supported only.")

    score_lines = numpy.genfromtxt(
        open_file(filename, mode="rb"),
        dtype=None,
        names=names,
        converters=converters,
        invalid_raise=True,
        usecols=usecols,
        **kwargs,
    )
    new_dtype = []
    for name in score_lines.dtype.names[:-1]:
        new_dtype.append((name, str(score_lines.dtype[name]).replace("S", "U")))
    new_dtype.append(("score", float))
    score_lines = numpy.array(score_lines, new_dtype)
    return score_lines


def load_files(filenames, func_load):
    """Load a list of score files and return a list of tuples of (neg, pos)

    Parameters
    ----------

    filenames : :any:`list`
        list of file paths
    func_load :
        function that can read files in the list

    Returns
    -------

    :any:`list`: [(neg,pos)] A list of tuples, where each tuple contains the
    ``negative`` and ``positive`` sceach system/probee.

    """
    if filenames is None:
        return None
    res = []
    for filepath in filenames:
        res.append(func_load(filepath))
    return res


def get_negatives_positives(score_lines):
    """Take the output of load_score and return negatives and positives.  This
    function aims to replace split_four_column and split_five_column but takes a
    different input. It's up to you to use which one.
    """

    pos_mask = score_lines["claimed_id"] == score_lines["real_id"]
    positives = score_lines["score"][pos_mask]
    negatives = score_lines["score"][numpy.logical_not(pos_mask)]
    return (negatives, positives)


def get_negatives_positives_from_file(filename, **kwargs):
    """Loads the scores first efficiently and then calls
    get_negatives_positives"""
    score_lines = load_score(filename, minimal=True, **kwargs)
    return get_negatives_positives(score_lines)


def get_negatives_positives_all(score_lines_list):
    """Take a list of outputs of load_score and return stacked negatives and
    positives.
    """

    negatives, positives = [], []
    for score_lines in score_lines_list:
        neg_pos = get_negatives_positives(score_lines)
        negatives.append(neg_pos[0])
        positives.append(neg_pos[1])
    negatives = numpy.vstack(negatives).T
    positives = numpy.vstack(positives).T
    return (negatives, positives)


def get_all_scores(score_lines_list):
    """Take a list of outputs of load_score and return stacked scores"""

    return numpy.vstack(
        [score_lines["score"] for score_lines in score_lines_list]
    ).T


def dump_score(filename, score_lines):
    """Dump scores that were loaded using :py:func:`load_score`
    The number of columns is automatically detected.
    """

    if len(score_lines.dtype) == 5:
        fmt = "%s %s %s %s %.9f"
    elif len(score_lines.dtype) == 4:
        fmt = "%s %s %s %.9f"
    else:
        raise ValueError("Only scores with 4 and 5 columns are supported.")
    numpy.savetxt(filename, score_lines, fmt=fmt)


def _estimate_score_file_format(filename, ncolumns=None):
    """Estimates the score file format from the given score file.
    If ``ncolumns`` is in ``(4,5)``, then ``ncolumns`` is returned instead.
    """
    if ncolumns in (4, 5):
        return ncolumns

    f = open_file(filename, "rb")
    try:
        line = f.readline()
        ncolumns = len(line.split())
    except Exception:
        logger.warn(
            "Could not guess the number of columns in file: {}. "
            "Assuming 4 column format.".format(filename)
        )
        ncolumns = 4
    finally:
        f.close()
    return ncolumns


def _iterate_score_file(filename):
    """Opens the score file for reading and yields the score file line by line in a tuple/list.

    The last element of the line (which is the score) will be transformed to float, the other elements will be str
    """
    if iscsv(filename):
        for row in _iterate_csv_score_file(filename):
            yield [
                row["bio_ref_subject_id"],
                row["probe_subject_id"],
                row["probe_key"],
                row["score"],
            ]
    else:
        opened = open_file(filename, "rb")
        import io

        if not isinstance(opened, io.TextIOWrapper):
            opened = io.TextIOWrapper(opened, newline="")

        reader = csv.reader(opened, delimiter=" ")
        for splits in reader:
            splits[-1] = float(splits[-1])
            yield splits


def _iterate_csv_score_file(filename):
    """Opens a CSV score file for reading and yields each line in a dict.

    The `score` field of the line will be cast to float, the other elements will
    be str.
    """
    opened = open_file(filename)
    reader = csv.DictReader(opened)
    for row in reader:
        row["score"] = float(row["score"])
        yield row


def _split_scores(
    score_lines, real_id_index, claimed_id_index=0, score_index=-1
):
    """Take the output of :py:func:`four_column` or :py:func:`five_column` and return negatives and positives."""
    positives, negatives = [], []
    for line in score_lines:
        which = (
            positives
            if line[claimed_id_index] == line[real_id_index]
            else negatives
        )
        which.append(line[score_index])

    return (numpy.array(negatives), numpy.array(positives))


def _split_cmc_scores(
    score_lines,
    real_id_index,
    probe_name_index=None,
    claimed_id_index=0,
    score_index=-1,
):
    """Takes the output of :py:func:`four_column` or :py:func:`five_column` and return cmc scores."""
    if probe_name_index is None:
        probe_name_index = real_id_index + 1
    # extract positives and negatives

    pos_dict = {}
    neg_dict = {}
    # read four column list
    for line in score_lines:
        which = (
            pos_dict
            if line[claimed_id_index] == line[real_id_index]
            else neg_dict
        )
        probe_name = line[probe_name_index]
        # append score
        if probe_name not in which:
            which[probe_name] = []
        which[probe_name].append(line[score_index])

    # convert to lists of tuples of ndarrays (or None)
    probe_names = sorted(set(neg_dict.keys()).union(set(pos_dict.keys())))
    # get all scores in the desired format
    return [
        (
            numpy.array(neg_dict[probe_name], numpy.float64)
            if probe_name in neg_dict
            else None,
            numpy.array(pos_dict[probe_name], numpy.float64)
            if probe_name in pos_dict
            else None,
        )
        for probe_name in probe_names
    ]


def split_csv_vuln(filename):
    """Loads vulnerability scores from a CSV score file.

    Returns the scores split between positive and negative as well as licit
    and presentation attack (spoof).

    The CSV must contain a `probe_attack_type` column with each field either
    containing a str defining the attack type (spoof), or empty (licit).

    Parameters
    ----------

    filename: str
        The path to a CSV file containing all the scores

    Returns
    -------

    split_scores: dict of str: numpy.ndarray
        The licit negative and positive, and spoof scores for probes.
    """
    logger.debug(f"Loading CSV score file: '{filename}'")
    split_scores = {"licit_neg": [], "licit_pos": [], "spoof": []}
    for row in _iterate_csv_score_file(filename):
        if not row["probe_attack_type"]:  # licit
            if row["probe_subject_id"] == row["bio_ref_subject_id"]:
                split_scores["licit_pos"].append(row["score"])
            else:
                split_scores["licit_neg"].append(row["score"])
        else:
            split_scores["spoof"].append(row["score"])
    logger.debug(
        f"Found {len(split_scores['licit_neg'])} negative (ZEI), "
        f"{len(split_scores['licit_pos'])} positive (licit), and "
        f"{len(split_scores['spoof'])} PA (spoof) scores."
    )
    # Cast to numpy float
    for key, val in split_scores.items():
        split_scores[key] = numpy.array(val, dtype=numpy.float64)
    return split_scores



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Tiago de Freitas Pereira <tiago.pereira@idiap.ch>


"""Executes biometric pipeline"""

import logging

import click

from clapper.click import ConfigCommand, ResourceOption, verbosity_option

from bob.pipelines.distributed import VALID_DASK_CLIENT_STRINGS

from .pipeline_simple import EPILOG as _SIMPLE_EPILOG

logger = logging.getLogger(__name__)


EPILOG = _SIMPLE_EPILOG.replace(
    "bob bio pipeline simple", "bob bio pipeline score-norm"
)


@click.command(
    name="score-norm",
    entry_point_group="bob.bio.config",
    cls=ConfigCommand,
    epilog=EPILOG,
)
@click.option(
    "--pipeline",
    "-p",
    entry_point_group="bob.bio.pipeline",
    required=True,
    help="PipelineSimple composed of a scikit-learn Pipeline and a BioAlgorithm",
    cls=ResourceOption,
)
@click.option(
    "--database",
    "-d",
    entry_point_group="bob.bio.database",
    required=True,
    help="Biometric Database connector (class that implements the methods: `background_model_samples`, `references` and `probes`)",
    cls=ResourceOption,
)
@click.option(
    "--dask-client",
    "-l",
    entry_point_group="dask.client",
    string_exceptions=VALID_DASK_CLIENT_STRINGS,
    default="single-threaded",
    help="Dask client for the execution of the pipeline.",
    cls=ResourceOption,
)
@click.option(
    "--group",
    "-g",
    "groups",
    type=click.Choice(["dev", "eval"]),
    multiple=True,
    default=("dev",),
    help="If given, this value will limit the experiments belonging to a particular protocolar group",
    cls=ResourceOption,
)
@click.option(
    "-o",
    "--output",
    show_default=True,
    default="results",
    help="Name of output directory where output scores will be saved. In case --checkpoint is set, checkpoints will be saved in this directory.",
    cls=ResourceOption,
)
@click.option(
    "--write-metadata-scores/--write-column-scores",
    "-m/-nm",
    default=True,
    help="If set, all the scores will be written with all their metadata using the `CSVScoreWriter`",
    cls=ResourceOption,
)
@click.option(
    "--memory",
    "-m",
    is_flag=True,
    help="If set, it will run the experiment keeping all objects on memory with nothing checkpointed. If not set, checkpoints will be saved in `--output`.",
    cls=ResourceOption,
)
@click.option(
    "--dask-partition-size",
    "-s",
    help="If using Dask, this option defines the size of each dask.bag.partition."
    "Use this option if the current heuristic that sets this value doesn't suit your experiment."
    "(https://docs.dask.org/en/latest/bag-api.html?highlight=partition_size#dask.bag.from_sequence).",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--dask-n-workers",
    "-n",
    help="If using Dask, this option defines the number of workers to start your experiment."
    "Dask automatically scales up/down the number of workers due to the current load of tasks to be solved."
    "Use this option if the current amount of workers set to start an experiment doesn't suit you.",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "-c",
    "--checkpoint-dir",
    show_default=True,
    default=None,
    help="Name of output directory where the checkpoints will be saved. In case --checkpoint is set, checkpoints will be saved in this directory.",
    cls=ResourceOption,
)
@click.option(
    "--top-norm",
    "-t",
    is_flag=True,
    help="If set, it will do the top-norm.",
    cls=ResourceOption,
)
@click.option(
    "--top-norm-score-fraction",
    default=1.0,
    type=float,
    help="Sets the percentage of samples used for t-norm and z-norm. Sometimes you don't want to use all the t/z samples for normalization",
    cls=ResourceOption,
)
@click.option(
    "--score-normalization-type",
    "-nt",
    type=click.Choice(["znorm", "tnorm"]),
    multiple=False,
    default="znorm",
    help="Type of normalization",
    cls=ResourceOption,
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="If set, it will force generate all the checkpoints of an experiment. This option doesn't work if `--memory` is set",
    cls=ResourceOption,
)
@click.option(
    "--no-dask",
    is_flag=True,
    help="If set, it will not use Dask to run the experiment.",
    cls=ResourceOption,
)
@verbosity_option(cls=ResourceOption, logger=logger)
def pipeline_score_norm(
    pipeline,
    database,
    dask_client,
    groups,
    output,
    write_metadata_scores,
    memory,
    dask_partition_size,
    dask_n_workers,
    checkpoint_dir,
    top_norm,
    top_norm_score_fraction,
    score_normalization_type,
    force,
    no_dask,
    **kwargs,
):
    """Runs the PipelineSimple with score normalization strategies

    Such pipeline consists into two major components.
    The first component consists of a scikit-learn `Pipeline`,
    where a sequence of transformations of the input data
    is defined.
    The second component is a `BioAlgorithm` that defines the primitives
    `enroll` and `score`

    With those two components any Biometric Experiment can be done.
    A Biometric experiment consists of three sub-pipelines and
    they are defined below:

    Sub-pipeline 1:\n
    ---------------

    Training background model.
    Some biometric algorithms demands the training of background model, for instance a neural network.

    \b
    This pipeline runs: `Pipeline.fit(DATA_FOR_FIT)`



    \b

    Sub-pipeline 2:\n
    ---------------

    Creation of biometric references: This is a standard step in a biometric pipelines.
    Given a set of samples of one identity, create a biometric reference (a.k.a template) for sub identity.


    \b
    raw_data --> preprocessing >> feature extraction >> enroll(background_model) --> biometric_reference

    This pipeline runs: `BioAlgorithm.enroll(Pipeline.transform(DATA_ENROLL))` >> biometric_references


    Sub-pipeline 3:\n
    ---------------

    Probing: This is another standard step in biometric pipelines.
    Given one sample and one biometric reference, computes a score.
    Such score has different meanings depending on the scoring method your biometric algorithm uses.
    It's out of scope to explain in a help message to explain what scoring is for different biometric algorithms.

    This pipeline runs: `BioAlgorithm.score(Pipeline.transform(DATA_SCORE, biometric_references))` >> biometric_references

    """
    from bob.bio.base.pipelines.entry_points import execute_pipeline_score_norm

    logger.debug("Executing PipelineScoreNorm")

    if no_dask:
        dask_client = None

    checkpoint = not memory

    execute_pipeline_score_norm(
        pipeline,
        database,
        dask_client,
        groups,
        output,
        write_metadata_scores,
        checkpoint,
        dask_partition_size,
        dask_n_workers,
        checkpoint_dir,
        top_norm,
        top_norm_score_fraction,
        score_normalization_type,
        force,
    )

    logger.info("Experiment finished !")



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :


import logging
import os

from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Optional, Union

import numpy as np

from sklearn.base import BaseEstimator

from bob.pipelines import Sample, SampleBatch, SampleSet
from bob.pipelines.wrappers import _frmt

logger = logging.getLogger(__name__)


def reduce_scores(
    scores: np.ndarray,
    axis: int,
    fn: Union[str, Callable[[np.ndarray, int], np.ndarray]] = "max",
):
    """
    Reduce scores using a function.

    Parameters:
    -----------
    scores
        Scores to reduce.

    fn
        Function to use for reduction. You can also provide a string like
        ``max`` to use the corresponding function from numpy. Some possible
        values are: ``max``, ``min``, ``mean``, ``median``, ``sum``.

    Returns:
    --------
    Reduced scores.
    """
    if isinstance(fn, str):
        fn = getattr(np, fn)
    return fn(scores, axis=axis)


def _data_valid(data: Any) -> bool:
    """Check if data is valid.

    Parameters:
    -----------
    data
        Data to check.

    Returns:
    --------
    True if data is valid, False otherwise.
    """
    if data is None:
        return False
    if isinstance(data, np.ndarray):
        return data.size > 0
    # we also have to check for [[]]
    if isinstance(data, list) and len(data) > 0:
        if isinstance(data[0], (list, tuple)):
            return len(data[0]) > 0
    return bool(data)


class BioAlgorithm(BaseEstimator, metaclass=ABCMeta):
    """Describes a base biometric comparator for the PipelineSimple
    :ref:`bob.bio.base.biometric_algorithm`.

    A biometric algorithm converts each SampleSet (which is a list of
    samples/features) into a single template. Template creation is done for both
    enroll and probe samples but the format of the templates can be different
    between enrollment and probe samples. After the creation of the templates,
    the algorithm computes one similarity score for comparison of an enroll
    template with a probe template.

    Examples
    --------
    >>> import numpy as np
    >>> from bob.bio.base.pipelines import BioAlgorithm
    >>> class MyAlgorithm(BioAlgorithm):
    ...
    ...     def create_templates(self, list_of_feature_sets, enroll):
    ...         # you cannot call np.mean(list_of_feature_sets, axis=1) because the
    ...         # number of features in each feature set may vary.
    ...         return [np.mean(feature_set, axis=0) for feature_set in list_of_feature_sets]
    ...
    ...     def compare(self, enroll_templates, probe_templates):
    ...         scores = []
    ...         for enroll_template in enroll_templates:
    ...             scores.append([])
    ...             for probe_template in probe_templates:
    ...                 similarity = 1 / np.linalg.norm(model - probe)
    ...                 scores[-1].append(similarity)
    ...         scores = np.array(scores, dtype=float)
    ...         return scores
    """

    def __init__(
        self,
        probes_score_fusion: Union[
            str, Callable[[list[np.ndarray], int], np.ndarray]
        ] = "max",
        enrolls_score_fusion: Union[
            str, Callable[[list[np.ndarray], int], np.ndarray]
        ] = "max",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.probes_score_fusion = probes_score_fusion
        self.enrolls_score_fusion = enrolls_score_fusion

    def fuse_probe_scores(self, scores, axis):
        return reduce_scores(scores, axis, self.probes_score_fusion)

    def fuse_enroll_scores(self, scores, axis):
        return reduce_scores(scores, axis, self.enrolls_score_fusion)

    @abstractmethod
    def create_templates(
        self, list_of_feature_sets: list[Any], enroll: bool
    ) -> list[Sample]:
        """Creates enroll or probe templates from multiple sets of features.

        The enroll template format can be different from the probe templates.

        Parameters
        ----------
        list_of_feature_sets
            A list of list of features with the shape of Nx?xD. N templates
            should be computed. Note that you cannot call
            np.array(list_of_feature_sets) because the number of features per
            set can be different depending on the database.
        enroll
            If True, the features are for enrollment. If False, the features are
            for probe.

        Returns
        -------
        templates
            A list of templates which has the same length as
            ``list_of_feature_sets``.
        """
        pass

    @abstractmethod
    def compare(
        self, enroll_templates: list[Sample], probe_templates: list[Sample]
    ) -> np.ndarray:
        """Computes the similarity score between all enrollment and probe templates.

        Parameters
        ----------
        enroll_templates
            A list (length N) of enrollment templates.

        probe_templates
            A list (length M) of probe templates.

        Returns
        -------
        scores
            A matrix of shape (N, M) containing the similarity scores.
        """
        pass

    def create_templates_from_samplesets(
        self, list_of_samplesets: list[SampleSet], enroll: bool
    ) -> list[Sample]:
        """Creates enroll or probe templates from multiple SampleSets.

        Parameters
        ----------
        list_of_samplesets
            A list (length N) of SampleSets.

        enroll
            If True, the SampleSets are for enrollment. If False, the SampleSets
            are for probe.

        Returns
        -------
        templates
            A list of Samples which has the same length as ``list_of_samplesets``.
            Each Sample contains a template.
        """
        logger.debug(
            f"{_frmt(self)}.create_templates_from_samplesets(... enroll={enroll})"
        )
        # create templates from .data attribute of samples inside sample_sets
        list_of_feature_sets = []
        for sampleset in list_of_samplesets:
            data = [s.data for s in sampleset.samples]
            valid_data = [d for d in data if d is not None]
            if len(data) != len(valid_data):
                logger.warning(
                    f"Removed {len(data)-len(valid_data)} invalid enrollment samples."
                )
            if not valid_data and enroll:
                # we do not support failure to enroll cases currently
                raise NotImplementedError(
                    f"None of the enrollment samples were valid for {sampleset}."
                )
            list_of_feature_sets.append(valid_data)

        templates = self.create_templates(list_of_feature_sets, enroll)
        expected_size = len(list_of_samplesets)
        assert len(templates) == expected_size, (
            "The number of (%s) templates (%d) created by the algorithm does not match "
            "the number of sample sets (%d)"
            % (
                "enroll" if enroll else "probe",
                len(templates),
                expected_size,
            )
        )
        # return a list of Samples (one per template)
        templates = [
            Sample(t, parent=sampleset)
            for t, sampleset in zip(templates, list_of_samplesets)
        ]
        return templates

    def score_sample_templates(
        self,
        probe_samples: list[Sample],
        enroll_samples: list[Sample],
        score_all_vs_all: bool,
    ) -> list[SampleSet]:
        """Computes the similarity score between all probe and enroll templates.

        Parameters
        ----------
        probe_samples
            A list (length N) of Samples containing probe templates.

        enroll_samples
            A list (length M) of Samples containing enroll templates.

        score_all_vs_all
            If True, the similarity scores between all probe and enroll templates
            are computed. If False, the similarity scores between the probes and
            their associated enroll templates are computed.

        Returns
        -------
        score_samplesets
            A list of N SampleSets each containing a list of M score Samples if score_all_vs_all
            is True. Otherwise, a list of N SampleSets each containing a list of <=M score Samples
            depending on the database.
        """
        logger.debug(
            f"{_frmt(self)}.score_sample_templates(... score_all_vs_all={score_all_vs_all})"
        )
        # Returns a list of SampleSets where a Sampleset for each probe
        # SampleSet where each Sample inside the SampleSets contains the score
        # for one enroll SampleSet
        score_samplesets = []
        if score_all_vs_all:
            probe_data = [s.data for s in probe_samples]
            valid_probe_indices = [
                i for i, d in enumerate(probe_data) if _data_valid(d)
            ]
            valid_probe_data = [probe_data[i] for i in valid_probe_indices]
            scores = self.compare(SampleBatch(enroll_samples), valid_probe_data)
            scores = np.asarray(scores, dtype=float)

            if len(valid_probe_indices) != len(probe_data):
                # inject None scores for invalid probe samples
                scores: list = scores.T.tolist()
                for i in range(len(probe_data)):
                    if i not in valid_probe_indices:
                        scores.insert(i, [None] * len(enroll_samples))
                # transpose back to original shape
                scores = np.array(scores, dtype=float).T

            expected_shape = (len(enroll_samples), len(probe_samples))
            assert scores.shape == expected_shape, (
                "The shape of the similarity scores (%s) does not match the expected shape (%s)"
                % (scores.shape, expected_shape)
            )
            for j, probe in enumerate(probe_samples):
                samples = []
                for i, enroll in enumerate(enroll_samples):
                    samples.append(Sample(scores[i, j], parent=enroll))
                score_samplesets.append(SampleSet(samples, parent=probe))
        else:
            for probe in probe_samples:
                references = [str(ref) for ref in probe.references]
                # get the indices of references for enroll samplesets
                indices = [
                    i
                    for i, enroll in enumerate(enroll_samples)
                    if str(enroll.template_id) in references
                ]
                if not indices:
                    raise ValueError(
                        f"No enroll sampleset found for probe {probe} and its required references {references}. "
                        "Did you mean to set score_all_vs_all=True?"
                    )
                if not _data_valid(probe.data):
                    scores = [[None]] * len(indices)
                else:
                    scores = self.compare(
                        SampleBatch([enroll_samples[i] for i in indices]),
                        SampleBatch([probe]),
                    )
                scores = np.asarray(scores, dtype=float)
                expected_shape = (len(indices), 1)
                assert scores.shape == expected_shape, (
                    "The shape of the similarity scores (%s) does not match the expected shape (%s)"
                    % (scores.shape, expected_shape)
                )
                samples = []
                for i, j in enumerate(indices):
                    samples.append(
                        Sample(scores[i, 0], parent=enroll_samples[j])
                    )
                score_samplesets.append(SampleSet(samples, parent=probe))

        return score_samplesets


class Database(metaclass=ABCMeta):
    """Base class for PipelineSimple databases"""

    def __init__(
        self,
        protocol: Optional[str] = None,
        score_all_vs_all: bool = False,
        annotation_type: Optional[str] = None,
        fixed_positions: Optional[str] = None,
        memory_demanding: bool = False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        protocol
            Name of the database protocol to use.
        score_all_vs_all
            Wether to allow scoring of all the probes against all the references, or to
            provide a list ``references`` provided with each probes to indicate against
            which references it needs to be compared.
        annotation_type
            The type of annotation passed to the annotation loading function.
        fixed_positions
            The constant eyes positions passed to the annotation loading function.
            TODO why keep this face-related name here? Which one is it, too (position
            when annotations are missing, or ending position in the result image)?
            --> move this when the FaceCrop annotator is correctly implemented.
        memory_demanding
            Flag to indicate that this should not be loaded locally.
            TODO Where is it used?
        """
        super().__init__(**kwargs)
        if not hasattr(self, "protocol"):
            self.protocol = protocol
        self.score_all_vs_all = score_all_vs_all
        self.annotation_type = annotation_type
        self.fixed_positions = fixed_positions
        self.memory_demanding = memory_demanding

    def __str__(self) -> str:
        args = ", ".join(
            [
                "{}={}".format(k, v)
                for k, v in self.__dict__.items()
                if not k.startswith("_")
            ]
        )
        return f"{self.__class__.__name__}({args})"

    @abstractmethod
    def background_model_samples(self) -> list[Sample]:
        """Returns :any:`Sample`\ s to train a background model


        Returns
        -------
        samples
            List of samples for background model training.

        """  # noqa: W605
        pass

    @abstractmethod
    def references(self, group: str = "dev") -> list[SampleSet]:
        """Returns references to enroll biometric references


        Parameters
        ----------
        group
            Limits samples to this group


        Returns
        -------
        references
            List of samples for the creation of biometric references.

        """
        pass

    @abstractmethod
    def probes(self, group: str = "dev") -> list[SampleSet]:
        """Returns probes to score against enrolled biometric references


        Parameters
        ----------
        group
            Limits samples to this group


        Returns
        -------
        probes
            List of samples for the creation of biometric probes.

        """
        pass

    @abstractmethod
    def all_samples(self, groups: Optional[str] = None) -> list[Sample]:
        """Returns all the samples of the dataset

        Parameters
        ----------
        groups
            List of groups to consider (like 'dev' or 'eval'). If `None`, will
            return samples from all the groups.

        Returns
        -------
        samples
            List of all the samples of the dataset.
        """
        pass

    @abstractmethod
    def groups(self) -> list[str]:
        """Returns all the possible groups for the current protocol."""
        pass

    @abstractmethod
    def protocols(self) -> list[str]:
        """Returns all the possible protocols of the database."""
        pass

    def template_ids(self, group: str) -> list[Any]:
        """Returns the ``template_id`` attribute of each reference."""
        return [s.template_id for s in self.references(group=group)]


class ScoreWriter(metaclass=ABCMeta):
    """
    Defines base methods to read, write scores and concatenate scores
    for :any:`bob.bio.base.pipelines.BioAlgorithm`
    """

    def __init__(self, path, extension=".txt", **kwargs):
        super().__init__(**kwargs)
        self.path = path
        self.extension = extension

    @abstractmethod
    def write(self, sampleset, path):
        pass

    def post_process(self, score_paths, filename):
        def _post_process(score_paths, filename):
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, "w") as f:
                for path in score_paths:
                    with open(path) as f2:
                        f.writelines(f2.readlines())
            return filename

        import dask
        import dask.bag

        if isinstance(score_paths, dask.bag.Bag):
            all_paths = dask.delayed(list)(score_paths)
            return dask.delayed(_post_process)(all_paths, filename)
        return _post_process(score_paths, filename)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Amir Mohammadi <amir.mohammadi@idiap.ch>
# Wed 20 July 16:20:12 CEST 2016
#

"""
Very simple tests for Implementations
"""

from pathlib import Path

import bob.bio.base

from bob.bio.base.config.dummy.database import database as dummy_database
from bob.bio.base.database import CSVDatabase
from bob.pipelines import DelayedSample, SampleSet

DATA_DIR = Path(__file__).parent / "data"


def test_all_samples():
    all_samples = dummy_database.all_samples(groups=None)
    assert len(all_samples) == 400
    assert all([isinstance(s, DelayedSample) for s in all_samples])
    assert len(dummy_database.all_samples(groups=["train"])) == 200
    assert len(dummy_database.all_samples(groups=["dev"])) == 200
    assert len(dummy_database.all_samples(groups=[])) == 400


def test_atnt():
    database = bob.bio.base.load_resource(
        "atnt", "database", preferred_package="bob.bio.base"
    )
    train_set = database.background_model_samples()
    assert len(train_set) > 0
    assert isinstance(train_set[0], DelayedSample)
    references = database.references()
    assert len(references) > 0
    references_sset = references[0]
    assert isinstance(references_sset, SampleSet)
    assert hasattr(references_sset, "key")
    assert hasattr(references_sset, "subject_id")
    assert hasattr(references_sset, "template_id")
    references_sample = references_sset.samples[0]
    assert isinstance(references_sample, DelayedSample)
    assert hasattr(references_sample, "key")
    probes = database.probes()
    assert len(probes) > 0
    assert isinstance(probes[0], SampleSet)
    assert isinstance(probes[0].samples[0], DelayedSample)
    all_samples = database.all_samples()
    assert len(all_samples) > 0
    assert isinstance(all_samples[0], DelayedSample)


def test_metadata():
    local_protocol_definition_path = DATA_DIR / "example_csv_filelist"
    database = CSVDatabase(
        name="dummy_metadata",
        protocol="protocol_only_dev_metadata",
        dataset_protocols_path=local_protocol_definition_path,
        templates_metadata=["subject_metadata"],
    )
    references_sset = database.references()[0]
    assert hasattr(references_sset, "subject_metadata")
    references_sample = references_sset.samples[0]
    assert hasattr(references_sample, "sample_metadata")
    probes_sset = database.probes()[0]
    assert hasattr(probes_sset, "subject_metadata")
    probes_sample = probes_sset.samples[0]
    assert hasattr(probes_sample, "sample_metadata")



from random import random

from bob.bio.base.annotator import Callable, FailSafe


def simple_annotator(image_batch, **kwargs):
    all_annotations = []
    for image in image_batch:
        all_annotations.append(
            {
                "topleft": (0, 0),
                "bottomright": image.shape,
            }
        )
    return all_annotations


def moody_annotator(image_batch, **kwargs):
    all_annotations = simple_annotator(image_batch, **kwargs)
    for annot in all_annotations:
        if random() < 0.5:
            del annot["bottomright"]
    return all_annotations


def fail_annotator(image_batch, **kwargs):
    all_annotations = []
    for image in image_batch:
        all_annotations.append({})
    return all_annotations


annotator = FailSafe(
    [Callable(fail_annotator), Callable(simple_annotator)],
    required_keys=["topleft", "bottomright"],
)



from bob.bio.base.config.dummy.algorithm import algorithm  # noqa: F401
from bob.bio.base.config.dummy.database import database  # noqa: F401
from bob.bio.base.config.dummy.extractor import extractor  # noqa: F401
from bob.bio.base.config.dummy.preprocessor import preprocessor  # noqa: F401

zt_norm = True
verbose = 1
sub_directory = "test_dummy"
